import os
import sys

import matplotlib
import matplotlib.pyplot as plt

# needed for the imports of our packages
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
# Change the working directory to the script directory
script_dir = os.path.dirname(os.path.abspath(__file__))
os.chdir(script_dir)

from src import GSS
from src.gss import image_util

# save output to the demo directory
image_dir = os.path.join(script_dir)

###################################################################################################
# This demo will analyse the GSS dataset for Equal Wealth combined with the social space.
###################################################################################################


# the parameters that gave us the best results
parameters = ((0.08, 0.08), 2.4, 15.4, 0.1, 0.1)


# the GSS class can be initialized with the names of the two datasets used in the paper, but this can easily be extended
# for other opinion within the data set.

# the class has two additional parameters. 1) The first year that is to be analysed, here the default is None,
# which means the first year in the data set will be used. 2) The final year used for the analysis which we set to 2018
equal_wealth = GSS("eqwlth")

# This function will handle all the loading of the data and formatting everything into usable datastructures.
# It will redistribute the data and create the initial conditions for the model. For details see the documentation.
equal_wealth.data_preparation()

# This function runs the simulation for the specified parameters and calculates the Kl divergence between the resulting
# distributions. For details se the documentation
equal_wealth.data_analysis(*parameters)

# plot the kl divergence between simulation and data distribution
fig, ax = plt.subplots(figsize=(2.81, 2.3833))
ax = equal_wealth.plot_kl_divergence(ax=ax)
ax.tick_params(labelsize=10)
plt.show()

# now plot the snapshots of the social space for the year of 2000 for the distributed data and the simulated data

# First we set a few parameter for plotting
xlim = (-0.35, 0.35)
ylim = (-0.35, 0.35)
width = 2.18
colorbar = True

fig = plt.figure(figsize=(width, 2.3833))
plt.rc("font", size=10)

# we only need one plot for the first year since data and simulation coincide

# plot the dataset
fig = equal_wealth.plot_distributed_data(
    year=2000,
    radii=parameters[0],
    figure=fig,
    network_alpha=0.05,
    s=0.1 * matplotlib.rcParams["lines.markersize"] ** 2,
    colorbar=False,
    network_lw=0.5,
    hist_ylim=594,
    xlim=xlim,
    ylim=ylim,
)
fig.show()
plt.close()

# plot the results of the simulation
fig = plt.figure(figsize=(width, 2.3833))
plt.rc("font", size=10)
fig = equal_wealth.model.plot_with_marginals(
    timestep=2000 - equal_wealth.first_year,
    xlim=xlim,
    ylim=ylim,
    print_title=False,
    xlabel=r"democrat $\longleftrightarrow$ republican",
    ylabel=r"liberal $\longleftrightarrow$ conservative",
    filename=None,
    bins=8,
    opinion_label=r"Government should help $\longleftrightarrow$ Individual responsability",
    no_ticks=True,
    colorbar=colorbar,
    s=0.1 * matplotlib.rcParams["lines.markersize"] ** 2,
    network_alpha=0.1,
    fig=fig,
    network_lw=0.5,
    hist_ylim=244,
    node_linewidth=0,  # set the linewidth around each agent to zero
)
fig.show()
plt.close()
