import os
import sys

# needed for the imports of our packages
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
from src import Model
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

# Change the working directory to the script directory
script_dir = os.path.dirname(os.path.abspath(__file__))
os.chdir(script_dir)

# The model is defined using the constructor of the Model class.
# Additional arguments that can be initialzed using the constructor are documented in the source code.
# The agents are generated upon calling initialize_model().
# The agents can be viewed using plot(), which generates a Axes object of the matplotlib library.

model = Model(
    N=100,
    R=0.15,
    alpha=20,
    beta=20,
    sigma_op=0.01,
    sigma_sp=0.01,
    delta_t=0.01,
    seed=123,
)
model.initialize_model()
fig = plt.figure()
model.plot()
fig.show()

# The agents of the model are initialized uniformly in the area defined by init_xlim and init_ylim, which describe the
# lower and upper bound of the x and y axis. The default for both of these parameters is (-0.25, 0.25).

# It is also possible to initialize the agents from a matrix by using the input_df argument of initialize_model().
# In this case the spatial coordinates and the opinions were generated from normal distributions.

rng = np.random.default_rng(123)

x = rng.normal(0, 0.01, 100)
y = rng.normal(0, 0.01, 100)
opinions = rng.normal(0, 1, 100)
input_df = pd.DataFrame({"x": x, "y": y, "opinions": opinions})

# initialize the model
model.initialize_model(input_df=input_df)
fig = plt.figure()
model.plot()
fig.show()

# The model is simulated using simulate(n), where n is the number of discrete timesteps that simulated.
# The default value of n is 1.

model.simulate(100)

# The Model class tracks the state of the model for each timestep in the class variable results,
# which can be obtained using get_results(). results is a dictionary which separates different types of information.
# The Opinions, Cluster Labels and gridcell contain a matrix, where the columns correspond to the timesteps of the
# simulation and the agents of the model are represented as rows. The others contain a list of either matrices,
# lists or numbers, where the list index corresponds to the timesteps.

results = model.get_results()
print(f"The keys to the reults dictionary are {results.keys()}")
print("\n")

# Giving the timestep to plot() will generate a plot of the social space for that specific time step
# While Calling model.plot() will plot the last timestep.

fig = plt.figure()
model.plot(timestep=50)
fig.show()

# To see how we cna access the data we we get use the timestep 50 as an example.
coordinates_50 = results["Coordinates"][50]
coordinates_50.columns = ["x", "y"]
opinions_50 = results["Opinions"][50]
data_50 = pd.DataFrame(coordinates_50)
data_50["opinions"] = opinions_50
print(data_50.head())
print("\n")

## Clustering

# The model clusters the agents based on the x and y coordinates in each timestep using the HDBSCAN clustering algorithm,
# due to the class variables useHDBCSAN and spatial_clustering having True as default values.
# It is possible to use spatial and/or opinion information for the clustering, by setting the spatial_clustering and
# opinion_clustering accordingly.
# Setting useHDBSCAN to False, results in the usage of DBSCAN as a clustering algorithm.
# For DBSCAN the hyperparameters eps and minpoints are set to 0.04 and 4, respectively. It is possible to recalculate
# the estimated optimal value of eps in each timestep as the elbow of the k-distance graph, by setting *dynamic_eps*
# to True.
#
# The corresponding clusters can be visualized with plot() and the cluster labels are in the results dictionary.
# For more details see the documentation.

fig = plt.figure()
model.plot(50, show_clusters=True)
fig.show()

clusters = results["Cluster Labels"][50]
data_50["clusters"] = clusters
print("The first row")
print(data_50.head())

## Potential
# A potential landscape has also been implemented, which is a double well potential. It can be enabled in the model by
# setting *useDoubleWell* to True using set_useDoubleWell() or directly setting the attribute when defining the instance.

model.initialize_model()
model.set_useDoubleWell(True)
model.simulate(50)
fig = plt.figure()
model.plot(show_clusters=True, show_landscape=True, ylim=(-0.5, 0.5), xlim=(-0.8, 0.8))
fig.show()

# For more information, please look up the documentation
