import os
import sys
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# needed for the imports of our packages
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))

from src import ModelStatistics

# Change the working directory to the script directory
script_dir = os.path.dirname(os.path.abspath(__file__))
os.chdir(script_dir)

# The ModelStatistics class is meant to mirror the usage of the underlying Model class from the previous demo.
# It takes the same parameters of the Model class, but also requires the number of simulations that should be simulated
# for each run.

# set the Model Parameters
number_of_runs = 4
N = 100
R_op = 0.15
R_sp = 0.15
alpha = 20
beta = 40
sigma_op = 0.05
sigma_sp = 0.05
delta_t = 0.01
integration_steps = 250

# We will mostly use the default settings, i.e. not providing an input_df, But it is possible for some scenarios.
# Here we use a single dataframe to set the initial data of all runs to be the same.
# It is also possible to set individual initial conditions to be passed to the Model class by setting
# input_df = [input_df_1, ..., input_df_k], this could be is used if you want to use your own distribution for the
# initial data.

rng = np.random.default_rng(123)

x = rng.normal(0, 0.01, 100)
y = rng.normal(0, 0.01, 100)
opinions = rng.normal(0, 1, 100)
input_df = pd.DataFrame({"x": x, "y": y, "opinions": opinions})

# set up the model
ensemble = ModelStatistics(
    number_of_runs,
    N,
    (R_op, R_sp),
    alpha,
    beta,
    sigma_op,
    sigma_sp,
    delta_t,
    spatial_clustering=False,
    opinion_clustering=False,
    input_df=input_df,
)
# The default of the statistics instance will also initialize all the models unless otherwise specified.

# run the simulation and will by default use all available cores for parallelization, otherwise they can be explicitly
# set as in this example.
ensemble.simulate(integration_steps, n_processes=4)

# The class contains a list of models, in which all the data is stored and all the functionality of the Model class
# Access the individual runs, indexed indices are [0, ..., number_of_runs-1]
run = 2
model = ensemble.models[run]

# plot the specific realization
fig, ax = plt.subplots(2, 1)
# plot the agents' position in social_space
model.plot(ax=ax[0], timestep=integration_steps)

# plotting parameters
plotting_time_step = 200
# plot the agents' opinions
model.plot_opinions(ax=ax[1], timestep=plotting_time_step, print_title=False)
plt.show()

# Access the data of each model individually
opinions = ensemble.models[run].results["Opinions"][plotting_time_step]
coordinates = ensemble.models[run].results["Coordinates"][plotting_time_step]


# Or access the data all at once:
# The opinions are stored in a datframe, such that they are all stacked on top of each other
print(
    f' The shape of the datframe is now {ensemble.get_results()["Opinions"].shape},'
    f" i.e. (num_runs * N, integration steps +1)\n"
)

# In the same manner the Coordinates value contains a list of dataframes corresponding to the timestep.
# Each dataframe contains the coordinates of all simulations stacked on top of each other.
print(
    f' The shape of the datframe is now {ensemble.get_results()["Coordinates"][250].shape},'
    f" i.e. (num_runs * N, 2)\n"
)


# The plotting functionality is extended with the same syntax where possible. calling model.plot_opinions() for
# an instance of the Model class would plot a histogram of the opinions of a single run. The
fig, ax = plt.subplots()
# Plot the opinion distribution over all runs
ensemble.plot_opinions(ax=ax, timestep=plotting_time_step)
plt.show()
