"""
Created on Tue Apr 09 2024
@author: Soeren Nagel

This script will generate all the figures in "Coevolving networks for opinion and social dynamics
in agent-based models:
"""

import os
import sys

import matplotlib
import matplotlib.pyplot as plt

# needed for the imports of our packages
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))

from src import GSS
from src.gss import image_util


def main(test=False):
    # Change the working directory to the script directory
    script_dir = os.path.dirname(os.path.abspath(__file__))
    os.chdir(script_dir)
    # change for trying out things
    save = True
    image_dir = os.path.join(script_dir, "..", "Images", "Paper")
    if test:
        image_dir = os.path.join(script_dir, "..", "test")
    ###################################################################################################
    # GSS Data Equal Wealth
    ###################################################################################################
    xlim = (-0.35, 0.35)
    ylim = (-0.35, 0.35)

    # the parameters that gave us the best results
    parameters = ((0.08, 0.08), 2.4, 15.4, 0.1, 0.1)

    # set the years we want to plot the comparison for
    election_years = (1978, 2000, 2016)

    # run the analysis
    equal_wealth = GSS("eqwlth")
    equal_wealth.data_preparation()
    equal_wealth.data_analysis(*parameters)

    # plot the kl divergence between simulation and data distribution
    fig, ax = plt.subplots(figsize=(2.81, 2.3833))
    ax = equal_wealth.plot_kl_divergence(ax=ax)
    ax.tick_params(labelsize=10)
    plt.savefig(os.path.join(image_dir, f"KL_eqwlth.png"), dpi=300)
    print(f"Figure saved to {image_dir}")
    plt.savefig(os.path.join(image_dir, f"KL_eqwlth.svg"))
    print(f"Figure saved to {image_dir}")

    # now plot the snapshots of the social space for the election years
    for year in election_years:
        width = 2.81 if year == 1978 else 2.18
        colorbar = True if year == 1978 else False
        fig = plt.figure(figsize=(width, 2.3833))
        plt.rc("font", size=10)

        # we only need one plot for the first year since data and simulation coincide
        if year != election_years[0]:
            # plot the dataset
            equal_wealth.plot_distributed_data(
                year=year,
                radii=parameters[0],
                filename=os.path.join(image_dir, f"data_{year}.png"),
                figure=fig,
                network_alpha=0.015,
                s=0.1 * matplotlib.rcParams["lines.markersize"] ** 2,
                colorbar=False,
                network_lw=0.5,
                hist_ylim=594,
                xlim=xlim,
                ylim=ylim,
            )
            plt.close()

        # plot the results of the simulation
        fig = plt.figure(figsize=(width, 2.3833))
        plt.rc("font", size=10)
        equal_wealth.model.plot_with_marginals(
            timestep=year - equal_wealth.first_year,
            xlim=xlim,
            ylim=ylim,
            print_title=False,
            xlabel=r"democrat $\longleftrightarrow$ republican",
            ylabel=r"liberal $\longleftrightarrow$ conservative",
            filename=os.path.join(image_dir, f"simulation{year}.png"),
            bins=8,
            opinion_label=r"Government should help $\longleftrightarrow$ Individual responsability",
            no_ticks=True,
            colorbar=colorbar,
            s=0.1 * matplotlib.rcParams["lines.markersize"] ** 2,
            network_alpha=0.015,
            fig=fig,
            network_lw=0.5,
            hist_ylim=244,
            node_linewidth=0,
        )
        plt.close()
    plt.rcdefaults()

    # Now we need to combine the images
    paths = [
        os.path.join(image_dir, f"simulation{year}.png") for year in election_years
    ]
    paths.append(os.path.join(image_dir, f"KL_eqwlth.png"))
    paths.append(os.path.join(image_dir, f"data_{2000}.png"))
    paths.append(os.path.join(image_dir, f"data_{2016}.png"))
    image_util.combine_pngs(*paths, os.path.join(image_dir, "EqualWealth.png"))
    ###############################################################################################
    # GSS Data Help Sick
    ###############################################################################################
    xlim = (-0.35, 0.35)
    ylim = (-0.35, 0.35)

    # the parameters that gave us the best results
    parameters = ((0.15, 0.15), 2.72, 12.1, 0.05, 0.05)

    # set the years we want to plot the comparison for
    election_years = (1975, 2000, 2016)

    # run the analysis
    help_sick = GSS("helpsick")
    help_sick.data_preparation()
    help_sick.data_analysis(*parameters)

    # plot the kl divergence between simulation and data distribution
    fig, ax = plt.subplots(figsize=(2.81, 2.3833))
    ax = help_sick.plot_kl_divergence(ax=ax)
    ax.tick_params(labelsize=10)
    plt.savefig(os.path.join(image_dir, f"KL_helpsick.png"), dpi=300)
    print(f"Figure saved to {image_dir}")
    plt.savefig(os.path.join(image_dir, f"KL_helpsick.svg"))
    print(f"Figure saved to {image_dir}")

    # now plot the snapshots of the social space for the election years
    for year in election_years:
        width = 2.81 if year == 1975 else 2.234
        colorbar = True if year == 1975 else False
        fig = plt.figure(figsize=(width, 2.3833))
        plt.rc("font", size=10)

        # we only need one plot for the first year since data and simulation coincide
        if year != election_years[0]:
            # plot the dataset
            help_sick.plot_distributed_data(
                year=year,
                radii=parameters[0],
                filename=os.path.join(image_dir, f"data_{year}.png"),
                figure=fig,
                network_alpha=0.015,
                s=0.1 * matplotlib.rcParams["lines.markersize"] ** 2,
                colorbar=False,
                network_lw=0.5,
                hist_ylim=674,
                xlim=xlim,
                ylim=ylim,
            )
            plt.close()

        # plot the results of the simulation
        fig = plt.figure(figsize=(width, 2.3833))
        plt.rc("font", size=10)
        help_sick.model.plot_with_marginals(
            timestep=year - help_sick.first_year,
            xlim=xlim,
            ylim=ylim,
            print_title=False,
            xlabel=r"democrat $\longleftrightarrow$ republican",
            ylabel=r"liberal $\longleftrightarrow$ conservative",
            filename=os.path.join(image_dir, f"simulation{year}.png"),
            bins=8,
            opinion_label=r"Government should help $\longleftrightarrow$ Individual responsibility",
            no_ticks=True,
            colorbar=colorbar,
            s=0.1 * matplotlib.rcParams["lines.markersize"] ** 2,
            network_alpha=0.015,
            fig=fig,
            network_lw=0.5,
            hist_ylim=620,
            node_linewidth=0,
        )
    plt.rcdefaults()

    # Now we need to combine the images
    paths = [
        os.path.join(image_dir, f"simulation{year}.png") for year in election_years
    ]
    paths.append(os.path.join(image_dir, f"KL_helpsick.png"))
    paths.append(os.path.join(image_dir, f"data_{2000}.png"))
    paths.append(os.path.join(image_dir, f"data_{2016}.png"))
    image_util.combine_pngs(*paths, os.path.join(image_dir, "Help_Sick.png"))

    # ##################################################################################################
    # plot participant numbers of the GSS
    # #################################################################################################

    fig, ax = plt.subplots(figsize=(7, 3.5))
    help_sick.plot_gss_participants(ax=ax)
    equal_wealth.plot_gss_participants(ax=ax)
    ax.legend(fontsize=14)
    ax.tick_params(axis="both", labelsize=14)
    plt.savefig(os.path.join(image_dir, "participants.svg"))
    print(f"Figure saved to {image_dir}")
    plt.savefig(os.path.join(image_dir, "participants.png"))
    print(f"Figure saved to {image_dir}")

    ###################################################################################################################
    # plot comparison for all election years
    ###################################################################################################################
    # We start with the equal_wealth data
    xlim = (-0.35, 0.35)
    ylim = (-0.35, 0.35)
    parameters = ((0.08, 0.08), 2.4, 15.4, 0.1, 0.1)
    election_years = (1978, 1988, 1990, 1996, 2000, 2004, 2008, 2012, 2016)

    image_paths = []
    for year in election_years:
        fig = plt.figure(figsize=(2.81, 2.3833))
        plt.rc("font", size=10)

        # plot the dataset
        equal_wealth.plot_distributed_data(
            year,
            radii=parameters[0],
            filename=os.path.join(image_dir, f"data_{year}.png"),
            figure=fig,
            network_alpha=0.015,
            s=0.1 * matplotlib.rcParams["lines.markersize"] ** 2,
            colorbar=True,
            network_lw=0.5,
            xlim=xlim,
            ylim=ylim,
        )
        plt.close()

        fig = plt.figure(figsize=(2.234, 2.3833))
        plt.rc("font", size=10)
        equal_wealth.model.plot_with_marginals(
            timestep=year - equal_wealth.first_year,
            xlim=xlim,
            ylim=ylim,
            print_title=False,
            xlabel=r"democrat $\longleftrightarrow$ republican",
            ylabel=r"liberal $\longleftrightarrow$ conservative",
            filename=os.path.join(image_dir, f"simulation{year}.png"),
            bins=8,
            opinion_label=r"Government should help $\longleftrightarrow$ Individual responsibility",
            no_ticks=True,
            colorbar=False,
            s=0.1 * matplotlib.rcParams["lines.markersize"] ** 2,
            network_alpha=0.015,
            fig=fig,
            network_lw=0.5,
            hist_ylim=244,
            node_linewidth=0,
        )
        plt.close()
        image_util.combine_images_side_by_side(
            os.path.join(image_dir, f"data_{year}.png"),
            os.path.join(image_dir, f"simulation{year}.png"),
            os.path.join(image_dir, f"equal_wealth_{year}.png"),
        )
        os.remove(os.path.join(image_dir, f"data_{year}.png"))
        os.remove(os.path.join(image_dir, f"simulation{year}.png"))
        image_paths.append(os.path.join(image_dir, f"equal_wealth_{year}.png"))
    image_util.stack_images_vertically(
        image_paths, os.path.join(image_dir, f"equal_wealth_overview.png")
    )
    # remove the individual snapshots
    for image in image_paths:
        os.remove(image)

    # #################################################################################################################
    # we now create the same plots for the helpsick data
    xlim = (-0.35, 0.35)
    ylim = (-0.35, 0.35)
    parameters = ((0.15, 0.15), 2.72, 12.1, 0.05, 0.05)
    election_years = (1975, 1988, 1990, 1996, 2000, 2004, 2008, 2012, 2016)

    image_paths = []
    for year in election_years:
        fig = plt.figure(figsize=(2.81, 2.3833))
        plt.rc("font", size=10)

        # plot the dataset
        help_sick.plot_distributed_data(
            year,
            radii=parameters[0],
            filename=os.path.join(image_dir, f"data_{year}.png"),
            figure=fig,
            network_alpha=0.015,
            s=0.1 * matplotlib.rcParams["lines.markersize"] ** 2,
            colorbar=True,
            network_lw=0.5,
            xlim=xlim,
            ylim=ylim,
        )
        plt.close()

        fig = plt.figure(figsize=(2.234, 2.3833))
        plt.rc("font", size=10)

        # plot the results of the simulation
        help_sick.model.plot_with_marginals(
            timestep=year - help_sick.first_year,
            xlim=xlim,
            ylim=ylim,
            print_title=False,
            xlabel=r"democrat $\longleftrightarrow$ republican",
            ylabel=r"liberal $\longleftrightarrow$ conservative",
            filename=os.path.join(image_dir, f"simulation{year}.png"),
            bins=8,
            opinion_label=r"Government should help $\longleftrightarrow$ Individual responsability",
            no_ticks=True,
            colorbar=False,
            s=0.1 * matplotlib.rcParams["lines.markersize"] ** 2,
            network_alpha=0.015,
            fig=fig,
            network_lw=0.5,
            node_linewidth=0,
        )
        plt.close()
        image_util.combine_images_side_by_side(
            os.path.join(image_dir, f"data_{year}.png"),
            os.path.join(image_dir, f"simulation{year}.png"),
            os.path.join(image_dir, f"help_sick{year}.png"),
        )
        os.remove(os.path.join(image_dir, f"data_{year}.png"))
        os.remove(os.path.join(image_dir, f"simulation{year}.png"))
        image_paths.append(os.path.join(image_dir, f"help_sick{year}.png"))
    image_util.stack_images_vertically(
        image_paths, os.path.join(image_dir, f"help_sick_overview.png")
    )

    # remove the individual snapshots
    for image in image_paths:
        os.remove(image)
    plt.rcdefaults()

    # combine the two overvies side by side into one figure
    image_util.combine_images_side_by_side(
        os.path.join(image_dir, f"help_sick_overview.png"),
        os.path.join(image_dir, f"equal_wealth_overview.png"),
        os.path.join(image_dir, f"overview.png"),
    )

    # remove the individual overview figures
    os.remove(os.path.join(image_dir, f"help_sick_overview.png"))
    os.remove(os.path.join(image_dir, f"equal_wealth_overview.png"))


if __name__ == "__main__":
    main()
