"""
Created on Tue Apr 09 2024
@author: Soeren Nagel

This script will generate all the figures in "Coevolving networks for opinion and social dynamics
in agent-based models:
"""

import os
import sys

import numpy as np
import matplotlib.pyplot as plt

# needed for the imports of our packages
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))

from src import BifurcationAnalysis


def main(test=False):
    # Change the working directory to the script directory
    script_dir = os.path.dirname(os.path.abspath(__file__))
    os.chdir(script_dir)

    image_dir = os.path.join(script_dir, "..", "Images", "Paper")
    data_dir = os.path.join(script_dir, "..", "data", "DataBank")
    if test:
        data_dir = os.path.join(script_dir, "..", "test")
    # ##############################################################################################
    # Assortativity
    # ##############################################################################################
    N = 100
    alpha = 40
    beta = 10
    R_sp = 0.15
    sigma_op = 0.05
    sigma_sp = 0.05
    delta_t = 0.01
    timestep = 1000 if not test else 10  # less computational load for the test case
    iterations = 100 if not test else 5  # less computational load for the test case
    # set the opinion radii for the parameter sweep
    R_op = np.concatenate(
        [
            np.linspace(0.03, 0.1, 3),
            np.linspace(0.12, 0.18, 6),
            np.linspace(0.2, 0.3, 3),
            np.array([0.15]),
        ]
    )
    if test:
        # less computational load for the test case
        R_op = np.linspace(0.03, 0.3, 7)
    R_op.sort()
    R_op = np.round(R_op, 4)
    parameters = [N, R_op, R_sp, alpha, beta, sigma_op, sigma_sp]
    parameter_sweep = BifurcationAnalysis(
        iterations,
        parameters,
        delta_t,
        data_bank=data_dir,
    )
    # will only simulate if data does not already saved
    parameter_sweep.simulate(timestep, clear=True)
    # ------------------------------------------------------------------------------------------------
    # Histograms
    # set the parameters for which we also want histograms (given as index of R_OP)
    hist_list = np.array([4, 7, 10]) if not test else np.array([0, 1, 2])
    plotting_timestep = 1000 if not test else 8
    fig, ax = plt.subplots(1, 3, figsize=(7, 3.5))
    for i, num in enumerate(hist_list):
        stats = parameter_sweep.bifurctaion_statistics[num]
        stats.clear()
        stats.load_results_from_numpy(parameter_sweep.data_bank)
        ax[i].hist(
            stats.global_assortativity(time_step=plotting_timestep),
            color="tab:red",
            bins=np.linspace(-0.1, 1, 13),
        )
        stats.clear()
        ax[i].set_ylim((0, 60))
        ax[i].set_title(f"$R_{{op}}={R_op[num]}$", fontsize=20)
        ax[i].tick_params(axis="both", labelsize=20)
    ax[1].set_xlabel("Global Assortativity", fontsize=20)
    ax[0].set_ylabel("Counts", fontsize=20)
    plt.tight_layout()
    plt.savefig(os.path.join(image_dir, "bimodality.png"), dpi=200)
    plt.savefig(os.path.join(image_dir, "bimodality.svg"))
    # -------------------------------------------------------------------------------------------------
    # Create time series for the global assortativity and the result of the parameter sweep for several values of R_op
    figure = plt.figure(figsize=(7, 7))
    yaxes_ticks = [0, 0.5, 1]
    yaxes_ticklabels = [r"$" + str(tick) + r"$" for tick in yaxes_ticks]
    xaxes_ticks = [0.03, 0.075, 0.15, 0.225, 0.3]
    xaxes_ticklabels = [r"$" + str(tick) + r"$" for tick in xaxes_ticks]
    ax1 = plt.subplot(211)
    ax1.tick_params(axis="both", labelsize=20)
    ax1 = parameter_sweep.plot_assortativity_time_series(
        ax1, plot_stdw=True, max_timestep=plotting_timestep, parameter_subset=[0, 6, 12]
    )
    ax1.set_xlabel(r"$t$", fontsize=20)
    ax1.set_ylabel("Global Assortativity", fontsize=20)
    ax1.set_ylim((-0.1, 1.1))
    ax1.set_yticks(yaxes_ticks, labels=yaxes_ticklabels, fontsize=20)
    # -------------------------------------------------------------------------------------------------
    # results of parameter sweep
    markers = ["d" for _ in R_op]
    for i in hist_list:
        markers[i] = "x"
    markers[0] = "*"
    markers[6] = "*"
    markers[-1] = "*"
    marker_colors = ["k" for _ in R_op]
    marker_colors[0] = "tab:purple"
    marker_colors[6] = "tab:red"
    marker_colors[-1] = "tab:green"

    ax2 = plt.subplot(212)
    ax2.set_ylim((-0.1, 1.1))

    ax2.tick_params(axis="both", labelsize=16)
    ax2.set_xticks(xaxes_ticks, labels=xaxes_ticklabels, fontsize=20)
    ax2.set_yticks(yaxes_ticks, labels=yaxes_ticklabels, fontsize=20)

    parameter_sweep.plot_diversity_metrics(
        "R_{op}",
        ["global_assortativity"],
        ax=ax2,
        print_title=False,
        ylabel="Global Assortativity",
        marker=markers,
        marker_colors=marker_colors,
        std_error=True,
        timestep=plotting_timestep,
    )

    ax2.set_xlabel(r"$R_{op}$", fontsize=20)
    ax2.set_ylabel("Global Assortativity", fontsize=20)
    plt.tight_layout()

    plt.savefig(os.path.join(image_dir, "regimes.jpg"))
    plt.savefig(os.path.join(image_dir, "regimes.svg"))
    plt.close()


if __name__ == "__main__":
    main()
