"""
Created on Tue Apr 09 2024
@author: Soeren Nagel

This script will generate all the figures in "Coevolving networks for opinion and social dynamics
in agent-based models:
"""

import os
import sys

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

# needed for the imports of our packages
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
script_dir = os.path.dirname(os.path.abspath(__file__))
os.chdir(script_dir)

from src import Model


def main(test=False):

    # model parameters
    N = 100
    R = 0.15
    sigma_op = 0.05
    sigma_sp = 0.05
    delta_t = 0.01
    seed = 123  # 130
    steps = 250
    local_coupling = False
    alpha = 40
    beta = 10
    R_ops = [0.03, 0.3]

    # plotting parameters
    bins = 31
    xlim = (-0.8, 0.8)
    ylim = (-0.5, 0.5)

    main_axes_ticks = [-0.3, 0, 0.3]
    main_axes_ticklabels = [r"$" + str(tick) + r"$" for tick in main_axes_ticks]

    inset_width = "30%"
    inset_height = "30%"

    # Note that ticks and xlim can be in conflict! Start by setting ticks and tick labels to [] and
    # then find the desired xlim and set the ticks at the end.
    opinion_axes_ticks = [-0.5, 0, 0.5]
    opinion_axes_tick_labels = [r"$-0.5$", "", r"$0.5$"]
    opinion_xlim = (-0.8, 0.8)
    opinion_ylim = (0, 60)

    figure_format = "png"

    image_dir = folder = os.path.join(script_dir, "..", "Images", "Paper")
    if test:
        image_dir = os.path.join(script_dir, "..", "test")

    # ----------------------------------------------------------------------------------------
    for steps in [250, 1000]:
        model = Model(
            N,
            (R_ops[0], R),
            alpha,
            beta,
            sigma_op,
            sigma_sp,
            delta_t,
            seed=seed,
            spatial_clustering=False,
            opinion_clustering=False,
            normalize_with_neighbour_num=local_coupling,
            useDoublewell=True,
        )
        model.initialize_model()
        model.simulate(steps)
        figure = plt.figure(figsize=(7, 3.5))

        sax1 = plt.subplot(121)
        model.plot(
            print_title=False,
            xlim=xlim,
            ylim=ylim,
            show_network=True,
            colorbar=False,
            use_spatial_network=True,
            show_landscape=True,
            landscape_colorbar=False,
        )
        sax1.set_xlabel(r"($a$) $R_{{op}} = {} $".format(R_ops[0]), fontsize=20)
        sax1.tick_params(axis="both", labelsize=14)
        sax1.set_xticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
        sax1.set_yticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)

        sax2 = inset_axes(
            sax1,
            width=inset_width,
            height=inset_height,
            loc="upper center",
            borderpad=0,
        )
        model.plot_opinions(
            xlim=opinion_xlim,
            ylim=opinion_ylim,
            bins=np.linspace(-1, 1, bins),
            print_title=False,
            ax=sax2,
            density=False,
        )
        sax2.tick_params(axis="both", labelsize=14)
        sax2.set_xticks(opinion_axes_ticks, labels=opinion_axes_tick_labels)
        # sax2.set_yticklabels([-.5,.5])
        # sax2.yaxis.tick_right()
        # sax2.xaxis.tick_top()
        # --------------------------------------------------------------------------------------------------
        opinion_ylim = (0, 120)
        model = Model(
            N,
            (R_ops[1], R),
            alpha,
            beta,
            sigma_op,
            sigma_sp,
            delta_t,
            seed=seed,
            spatial_clustering=False,
            opinion_clustering=False,
            normalize_with_neighbour_num=local_coupling,
            useDoublewell=True,
        )
        model.initialize_model()
        model.simulate(steps)

        sax3 = plt.subplot(122)
        model.plot(
            print_title=False,
            xlim=xlim,
            ylim=ylim,
            show_network=True,
            colorbar_ticks=[-1, 0, 1],
            colorbar_fontsize=20,
            use_spatial_network=True,
            show_landscape=True,
            landscape_colorbar=False,
            colorbar=False,
        )
        sax3.tick_params(axis="both", labelsize=14)
        sax3.set_xlabel(r"($b$) $R_{{op}} = {} $".format(R_ops[1]), fontsize=20)
        sax3.set_yticks(main_axes_ticks, labels=[])
        sax3.set_xticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
        sax4 = inset_axes(
            sax3,
            width=inset_width,
            height=inset_height,
            loc="upper center",
            borderpad=0,
        )
        model.plot_opinions(
            xlim=opinion_xlim,
            ylim=opinion_ylim,
            bins=np.linspace(-1, 1, bins),
            print_title=False,
            ax=sax4,
            density=False,
        )
        sax4.tick_params(axis="both", labelsize=14)
        sax4.set_xticks(opinion_axes_ticks, opinion_axes_tick_labels)
        # sax4.yaxis.tick_right()
        # sax4.xaxis.tick_top()

        plt.tight_layout()
        plt.savefig(
            os.path.join(image_dir, f"Fig4_with_potential_T={delta_t * steps}.svg")
        )
        print(f"Figure saved to {image_dir}")

        plt.close()


if __name__ == "__main__":
    main()
