"""
Created on Tue Apr 09 2024
@author: Soeren Nagel

This script will generate all the figures in "Coevolving networks for opinion and social dynamics
in agent-based models:
"""

import os
import sys

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

# needed for the imports of our packages
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
script_dir = os.path.dirname(os.path.abspath(__file__))
os.chdir(script_dir)

from src import Model


def main(test=False):
    ###################################################################################################
    # Single Runs
    ###################################################################################################
    N = 100
    R = 0.15
    sigma_op = 0.05
    sigma_sp = 0.05
    delta_t = 0.01
    seed = 123  # 130
    steps = 250
    local_coupling = False

    bins = 31

    image_dir = "../Images/Paper/"
    if test:
        image_dir = os.path.join(script_dir, "..", "test")

    figure_format = "png"
    xlim = (-0.5, 0.65)
    ylim = (-0.5, 0.65)
    # =============================================================================
    # Further results
    # =============================================================================
    alpha = 40
    beta = 10
    R_sps = [0.03, 0.3]

    main_axes_ticks = [-0.3, 0, 0.3]
    main_axes_ticklabels = [r"$" + str(tick) + r"$" for tick in main_axes_ticks]

    inset_width = "30%"
    inset_height = "30%"

    # Note that ticks and xlim can be in conflict! Start by setting ticks and tick labels to [] and
    # then find the desired xlim and set the ticks at the end.
    opinion_axes_ticks = [-0.5, 0, 0.5]
    opinion_axes_tick_labels = [r"$-0.5$", "", r"$0.5$"]
    opinion_xlim = (-0.8, 0.8)
    opinion_ylim = (0, 60)
    # ----------------------------------------------------------------------------------------
    model = Model(
        N,
        (R, R_sps[0]),
        alpha,
        beta,
        sigma_op,
        sigma_sp,
        delta_t,
        seed=seed,
        spatial_clustering=False,
        opinion_clustering=False,
        normalize_with_neighbour_num=local_coupling,
    )
    model.initialize_model()
    model.simulate(steps)
    figure = plt.figure(figsize=(7, 3.5))

    sax1 = plt.subplot(121)
    model.plot(
        print_title=False, xlim=xlim, ylim=ylim, show_network=True, colorbar=False
    )
    sax1.set_xlabel(r"($a$) $R_{{sp}} = {} $".format(R_sps[0]), fontsize=20)
    sax1.tick_params(axis="both", labelsize=14)
    sax1.set_xticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
    sax1.set_yticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)

    sax2 = inset_axes(
        sax1, width=inset_width, height=inset_height, loc="upper right", borderpad=0
    )
    model.plot_opinions(
        xlim=opinion_xlim,
        ylim=opinion_ylim,
        bins=np.linspace(-1, 1, bins),
        print_title=False,
        ax=sax2,
        density=False,
    )
    sax2.tick_params(axis="both", labelsize=14)
    sax2.set_xticks(opinion_axes_ticks, labels=opinion_axes_tick_labels)
    # sax2.set_yticklabels([-.5,.5])
    # sax2.yaxis.tick_right()
    # sax2.xaxis.tick_top()
    # --------------------------------------------------------------------------------------------------
    opinion_ylim = (0, 120)
    model = Model(
        N,
        (R, R_sps[1]),
        alpha,
        beta,
        sigma_op,
        sigma_sp,
        delta_t,
        seed=seed,
        spatial_clustering=False,
        opinion_clustering=False,
        normalize_with_neighbour_num=local_coupling,
    )
    model.initialize_model()
    model.simulate(steps)

    sax3 = plt.subplot(122)
    model.plot(
        print_title=False,
        xlim=ylim,
        ylim=xlim,
        show_network=True,
        colorbar_ticks=[-1, 0, 1],
        colorbar_fontsize=20,
    )
    sax3.tick_params(axis="both", labelsize=14)
    sax3.set_xlabel(r"($b$) $R_{{sp}} = {} $".format(R_sps[1]), fontsize=20)
    sax3.set_yticks(main_axes_ticks, labels=[])
    sax3.set_xticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
    sax4 = inset_axes(
        sax3, width=inset_width, height=inset_height, loc="upper right", borderpad=0
    )
    model.plot_opinions(
        xlim=opinion_xlim,
        ylim=opinion_ylim,
        bins=np.linspace(-1, 1, bins),
        print_title=False,
        ax=sax4,
        density=False,
    )
    sax4.tick_params(axis="both", labelsize=14)
    sax4.set_xticks(opinion_axes_ticks, opinion_axes_tick_labels)
    # sax4.yaxis.tick_right()
    # sax4.xaxis.tick_top()

    plt.tight_layout()
    plt.savefig(os.path.join(image_dir, f"Fig1b-changing_R_sp.{figure_format}"))
    print(f"Figure saved to {image_dir}")

    plt.close()
    # =============================================================================
    # opinion radius
    # =============================================================================
    alpha = 40
    beta = 10
    R_ops = [0.03, 0.3]

    main_axes_ticks = [-0.3, 0, 0.3]
    main_axes_ticklabels = [r"$" + str(tick) + r"$" for tick in main_axes_ticks]

    inset_width = "30%"
    inset_height = "30%"

    # Note that ticks and xlim can be in conflict! Start by setting ticks and tick labels to [] and
    # then find the desired xlim and set the ticks at the end.
    opinion_axes_ticks = [-0.5, 0, 0.5]
    opinion_axes_tick_labels = [r"$-0.5$", "", r"$0.5$"]
    opinion_xlim = (-0.8, 0.8)
    opinion_ylim = (0, 60)
    # ----------------------------------------------------------------------------------------
    model = Model(
        N,
        (R_ops[0], R),
        alpha,
        beta,
        sigma_op,
        sigma_sp,
        delta_t,
        seed=seed,
        spatial_clustering=False,
        opinion_clustering=False,
        normalize_with_neighbour_num=local_coupling,
    )
    model.initialize_model()
    model.simulate(steps)
    figure = plt.figure(figsize=(7, 3.5))

    sax1 = plt.subplot(121)
    model.plot(
        print_title=False,
        xlim=xlim,
        ylim=ylim,
        show_network=True,
        colorbar=False,
        use_spatial_network=True,
    )
    sax1.set_xlabel(r"($a$) $R_{{op}} = {} $".format(R_ops[0]), fontsize=20)
    sax1.tick_params(axis="both", labelsize=14)
    sax1.set_xticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
    sax1.set_yticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)

    sax2 = inset_axes(
        sax1, width=inset_width, height=inset_height, loc="upper right", borderpad=0
    )
    model.plot_opinions(
        xlim=opinion_xlim,
        ylim=opinion_ylim,
        bins=np.linspace(-1, 1, bins),
        print_title=False,
        ax=sax2,
        density=False,
    )
    sax2.tick_params(axis="both", labelsize=14)
    sax2.set_xticks(opinion_axes_ticks, labels=opinion_axes_tick_labels)
    # sax2.set_yticklabels([-.5,.5])
    # sax2.yaxis.tick_right()
    # sax2.xaxis.tick_top()
    # --------------------------------------------------------------------------------------------------
    opinion_ylim = (0, 120)
    model = Model(
        N,
        (R_ops[1], R),
        alpha,
        beta,
        sigma_op,
        sigma_sp,
        delta_t,
        seed=seed,
        spatial_clustering=False,
        opinion_clustering=False,
        normalize_with_neighbour_num=local_coupling,
    )
    model.initialize_model()
    model.simulate(steps)

    sax3 = plt.subplot(122)
    model.plot(
        print_title=False,
        xlim=ylim,
        ylim=xlim,
        show_network=True,
        colorbar_ticks=[-1, 0, 1],
        colorbar_fontsize=20,
        use_spatial_network=False,
    )
    sax3.tick_params(axis="both", labelsize=14)
    sax3.set_xlabel(r"($b$) $R_{{op}} = {} $".format(R_ops[1]), fontsize=20)
    sax3.set_yticks(main_axes_ticks, labels=[])
    sax3.set_xticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
    sax4 = inset_axes(
        sax3, width=inset_width, height=inset_height, loc="upper right", borderpad=0
    )
    model.plot_opinions(
        xlim=opinion_xlim,
        ylim=opinion_ylim,
        bins=np.linspace(-1, 1, bins),
        print_title=False,
        ax=sax4,
        density=False,
    )
    sax4.tick_params(axis="both", labelsize=14)
    sax4.set_xticks(opinion_axes_ticks, opinion_axes_tick_labels)
    # sax4.yaxis.tick_right()
    # sax4.xaxis.tick_top()

    plt.tight_layout()
    plt.savefig(os.path.join(image_dir, f"R_op.{figure_format}"))
    print(f"Figure saved to {image_dir}")

    plt.close()
    # =============================================================================
    # alpha beta 1 to 5
    # =============================================================================
    alphas = [10, 100]
    betas = [50, 50]
    main_axes_ticks = [-0.3, 0, 0.3]
    main_axes_ticklabels = [r"$" + str(tick) + r"$" for tick in main_axes_ticks]
    xlim = (-0.5, 0.6)
    ylim = (-0.5, 0.6)

    inset_width = "30%"
    inset_height = "30%"
    opinion_axes_ticks = [-0.5, 0, 0.5]
    opinion_axes_tick_labels = [r"$-0.5$", "", r"$0.5$"]
    opinion_xlim = (-0.8, 0.8)
    opinion_ylim = (0, 60)
    # ----------------------------------------------------------------------------------------
    model = Model(
        N,
        R,
        alphas[0],
        betas[0],
        sigma_op,
        sigma_sp,
        delta_t,
        seed=seed,
        spatial_clustering=False,
        opinion_clustering=False,
        normalize_with_neighbour_num=local_coupling,
    )
    model.initialize_model()
    model.simulate(steps)

    figure = plt.figure(figsize=(7, 3.5))
    sax1 = plt.subplot(121)
    model.plot(
        print_title=False,
        xlim=xlim,
        ylim=ylim,
        show_network=False,
        colorbar=False,
    )
    sax1.set_xlabel(
        r"($a$) $\alpha = {} $ $\beta = {} $".format(alphas[0], betas[0]),
        fontsize=20,
    )
    sax1.tick_params(axis="both", labelsize=14)
    sax1.set_xticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
    sax1.set_yticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)

    sax2 = inset_axes(
        sax1, width=inset_width, height=inset_height, loc="upper right", borderpad=0
    )
    model.plot_opinions(
        xlim=opinion_xlim,
        ylim=opinion_ylim,
        bins=np.linspace(-1, 1, bins),
        print_title=False,
        ax=sax2,
        density=False,
    )
    sax2.tick_params(axis="both", labelsize=14)
    sax2.set_xticks(opinion_axes_ticks, labels=opinion_axes_tick_labels)
    # sax2.set_yticklabels([-.5,.5])
    # sax2.yaxis.tick_right()
    # sax2.xaxis.tick_top()
    # --------------------------------------------------------------------------------------------------
    model = Model(
        N,
        R,
        alphas[1],
        betas[1],
        sigma_op,
        sigma_sp,
        delta_t,
        seed=seed,
        spatial_clustering=False,
        opinion_clustering=False,
        normalize_with_neighbour_num=local_coupling,
    )
    model.initialize_model()
    model.simulate(steps)

    sax3 = plt.subplot(122)
    model.plot(
        print_title=False,
        xlim=xlim,
        ylim=ylim,
        show_network=False,
        colorbar_ticks=[-1, 0, 1],
        colorbar_fontsize=20,
    )
    sax3.tick_params(axis="both", labelsize=14)
    sax3.set_xlabel(
        r"($b$) $\alpha = {} $ $\beta = {} $".format(alphas[1], betas[1]),
        fontsize=20,
    )
    sax3.set_yticks(main_axes_ticks, labels=[])
    sax3.set_xticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
    sax4 = inset_axes(
        sax3, width=inset_width, height=inset_height, loc="upper right", borderpad=0
    )
    model.plot_opinions(
        xlim=opinion_xlim,
        ylim=opinion_ylim,
        bins=np.linspace(-1, 1, bins),
        print_title=False,
        ax=sax4,
        density=False,
    )
    sax4.tick_params(axis="both", labelsize=14)
    sax4.set_xticks(opinion_axes_ticks, opinion_axes_tick_labels)
    # sax4.yaxis.tick_right()
    # sax4.xaxis.tick_top()

    plt.tight_layout()
    plt.savefig(os.path.join(image_dir, f"alphabeta1-5.{figure_format}"))
    print(f"Figure saved to {image_dir}")
    plt.close()

    # =============================================================================
    # beta
    # =============================================================================
    alphas = [40, 40]
    betas = [10, 40]
    main_axes_ticks = [-0.3, 0, 0.3]
    main_axes_ticklabels = [r"$" + str(tick) + r"$" for tick in main_axes_ticks]
    xlim = (-0.5, 0.6)
    ylim = (-0.5, 0.6)

    inset_width = "30%"
    inset_height = "30%"
    opinion_axes_ticks = [-0.5, 0, 0.5]
    opinion_axes_tick_labels = [r"$-0.5$", "", r"$0.5$"]
    opinion_xlim = (-0.8, 0.8)
    opinion_ylim = (0, 120)
    # ----------------------------------------------------------------------------------------
    model = Model(
        N,
        R,
        alphas[0],
        betas[0],
        sigma_op,
        sigma_sp,
        delta_t,
        seed=seed,
        spatial_clustering=False,
        opinion_clustering=False,
        normalize_with_neighbour_num=local_coupling,
    )
    model.initialize_model()
    model.simulate(steps)
    figure = plt.figure(figsize=(7, 3.5))

    sax1 = plt.subplot(121)
    model.plot(
        print_title=False,
        xlim=xlim,
        ylim=ylim,
        show_network=False,
        colorbar=False,
    )
    sax1.set_xlabel(
        r"($a$) $\alpha = {} $ $\beta = {} $".format(alphas[0], betas[0]),
        fontsize=20,
    )
    sax1.tick_params(axis="both", labelsize=14)
    sax1.set_xticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
    sax1.set_yticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)

    sax2 = inset_axes(
        sax1, width=inset_width, height=inset_height, loc="upper right", borderpad=0
    )
    model.plot_opinions(
        xlim=opinion_xlim,
        ylim=opinion_ylim,
        bins=np.linspace(-1, 1, bins),
        print_title=False,
        ax=sax2,
        density=False,
    )
    sax2.tick_params(axis="both", labelsize=14)
    sax2.set_xticks(opinion_axes_ticks, labels=opinion_axes_tick_labels)
    # sax2.set_yticklabels([-.5,.5])
    # sax2.yaxis.tick_right()
    # sax2.xaxis.tick_top()
    # --------------------------------------------------------------------------------------------------
    opinion_ylim = (0, 60)
    model = Model(
        N,
        R,
        alphas[1],
        betas[1],
        sigma_op,
        sigma_sp,
        delta_t,
        seed=seed,
        spatial_clustering=False,
        opinion_clustering=False,
        normalize_with_neighbour_num=local_coupling,
    )
    model.initialize_model()
    model.simulate(steps)

    sax3 = plt.subplot(122)
    model.plot(
        print_title=False,
        xlim=xlim,
        ylim=ylim,
        show_network=False,
        colorbar_ticks=[-1, 0, 1],
        colorbar_fontsize=20,
    )
    sax3.tick_params(axis="both", labelsize=14)
    sax3.set_xlabel(
        r"($b$) $\alpha = {} $ $\beta = {} $".format(alphas[1], betas[1]),
        fontsize=20,
    )
    sax3.set_yticks(main_axes_ticks, labels=[])
    sax3.set_xticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
    sax4 = inset_axes(
        sax3, width=inset_width, height=inset_height, loc="upper right", borderpad=0
    )
    model.plot_opinions(
        xlim=opinion_xlim,
        ylim=opinion_ylim,
        bins=np.linspace(-1, 1, bins),
        print_title=False,
        ax=sax4,
        density=False,
    )
    sax4.tick_params(axis="both", labelsize=14)
    sax4.set_xticks(opinion_axes_ticks, opinion_axes_tick_labels)
    # sax4.yaxis.tick_right()
    # sax4.xaxis.tick_top()

    plt.tight_layout()
    plt.savefig(os.path.join(image_dir, f"beta.{figure_format}"))
    print(f"Figure saved to {image_dir}")
    plt.close()

    # =============================================================================
    # transient dynamics
    # =============================================================================
    xlim = (-0.55, 0.35)
    ylim = (-0.4, 0.4)
    main_axes_ticks = [-0.3, 0, 0.3]
    main_axes_ticklabels = [r"$" + str(tick) + r"$" for tick in main_axes_ticks]
    timestep = 450 if not test else 10

    # --------------------------------------------------------------------------------------------------
    model = Model(
        N,
        R,
        40,
        10,
        sigma_op,
        sigma_sp,
        delta_t,
        seed=seed,
        spatial_clustering=True,
        opinion_clustering=False,
        normalize_with_neighbour_num=local_coupling,
    )
    model.initialize_model()

    model.simulate(timestep)
    t = [0.55, 1.2, 1.6, 1.9, 2.5, 3.4]
    if test:
        t = [0.01, 0.02, 0.04, 0.06, 0.07, 0.08]
    figure = plt.figure(figsize=(7, 7))

    sax1 = plt.subplot(221)
    model.plot(
        timestep=int(t[0] * 100),
        ax=sax1,
        show_clusters=True,
        print_title=False,
        show_network=True,
        colorbar=False,
        xlim=xlim,
        ylim=ylim,
        cluster_legend_outside=False,
    )
    sax1.set_xticks(main_axes_ticks, labels=[])
    sax1.set_yticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
    sax1.set_xlabel(r"($a$) $T={}$ ".format(t[0]), fontsize=20)

    sax2 = plt.subplot(222)
    model.plot(
        timestep=int(t[1] * 100),
        ax=sax2,
        show_clusters=True,
        print_title=False,
        show_network=True,
        colorbar=False,
        xlim=xlim,
        ylim=ylim,
        cluster_legend_outside=False,
    )
    sax2.set_xlabel(r"($b$) $T={}$ ".format(t[1]), fontsize=20)
    sax2.set_xticks(main_axes_ticks, labels=[])
    sax2.set_yticks(main_axes_ticks, labels=[])

    sax3 = plt.subplot(223)
    model.plot(
        timestep=int(t[3] * 100),
        ax=sax3,
        show_clusters=True,
        print_title=False,
        show_network=True,
        colorbar=False,
        xlim=xlim,
        ylim=ylim,
        cluster_legend_outside=False,
    )
    sax3.tick_params(axis="both", labelsize=13)
    sax3.set_xlabel(r"($c$) $T={}$ ".format(t[3]), fontsize=20)
    sax3.set_xticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
    sax3.set_yticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)

    sax4 = plt.subplot(224)
    model.plot(
        timestep=int(t[5] * 100),
        ax=sax4,
        show_clusters=True,
        print_title=False,
        show_network=True,
        colorbar_ticks=[-1, 0, 1],
        colorbar_fontsize=20,
        xlim=xlim,
        ylim=ylim,
        cluster_legend_outside=False,
    )
    sax4.set_xticks(main_axes_ticks, labels=main_axes_ticklabels, fontsize=20)
    sax4.set_yticks(main_axes_ticks, labels=[])
    sax4.set_xlabel(r"($d$) $T={}$ ".format(t[5]), fontsize=20)

    plt.tight_layout()

    plt.savefig(os.path.join(image_dir, f"Network_Dynaimcs.{figure_format}"))
    print(f"Figure saved to {image_dir}")
    plt.close()


if __name__ == "__main__":
    main()
