from typing import Optional
import warnings

import joblib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from src import ModelStatistics


class BifurcationAnalysis:
    """
    This class will handle all the bifurcation analysis and parameter variation. It is structured the same way as the
    ModelStatistics class.
    """

    def __init__(
        self,
        num_iterations: int,
        parameters: list,
        delta_t: float,
        init_xlim: tuple[float, float] = (-0.25, 0.25),
        init_ylim: tuple[float, float] = (-0.25, 0.25),
        init_oplim: tuple[float, float] = (-1, 1),
        useDoublewell: bool = False,
        spatial_clustering=True,
        opinion_clustering: bool = True,
        dynamic_eps: bool = False,
        useHDBSCAN: bool = True,
        datasim: bool = False,
        xlim: Optional[tuple[float, float]] = None,
        ylim: Optional[tuple[float, float]] = None,
        boundary: bool = True,
        normalize_with_neighbour_num: bool = False,
        input_df: Optional[pd.DataFrame] = None,
        initialize: bool = True,
        data_bank: str | None = None,
    ) -> None:
        """

        @param num_iterations: int
            number of iterations performed for each set of parameter set
        @param parameters: a nested list of parameters
            A list with all the parameters is given,
            where all the fixed parameters are given as int ot float and
            the parameter for which we perform the bifurcation analysis given as a list or numpy array instead.
        @param delta_t:
        @param init_xlim:
        @param init_ylim:
        @param init_oplim:
        @param useDoublewell:
        @param spatial_clustering:
        @param opinion_clustering:
        @param dynamic_eps:
        @param useHDBSCAN:
        @param datasim:
        @param xlim:
        @param ylim:
        @param boundary:
        @param normalize_with_neighbour_num:
        @param input_df:
        @param initialize:
        """
        parameter_strings = (
            r"N",
            r"R_{op}",
            r"R_{sp}",
            r"\alpha",
            r"\beta",
            r"\sigma_{op}",
            r"\sigma_{sp}",
        )
        self.bifurcation_parameters = {}

        # finding the bifurcation parameters
        for i, parameter in enumerate(parameters):
            if isinstance(parameter, (list, np.ndarray)):
                self.bifurcation_parameters[parameter_strings[i]] = parameter

        # ensure the bifurcation parameters have the same length. i.e. define a curve through parameter space
        assert len(set([len(par) for par in self.bifurcation_parameters.values()])) == 1
        self.num_parameters = len(next(iter(self.bifurcation_parameters.values())))

        for i, parameter in enumerate(parameters):
            if parameter_strings[i] not in self.bifurcation_parameters.keys():
                parameters[i] = parameter * np.ones(self.num_parameters)

        # Now self.parameters[i] gives the relevant parameters for the i-th step of the bifurcations
        self.parameters = np.array(parameters).T

        self.delta_t = delta_t
        self.xlim = init_xlim
        self.ylim = init_ylim
        self.oplim = init_oplim
        self.useDoublewell = useDoublewell
        self.spatial_clustering = spatial_clustering
        self.opinion_clustering = opinion_clustering
        self.dynamic_eps = dynamic_eps
        self.useHDBSCAN = useHDBSCAN
        self.datasim = datasim
        self.xlim = xlim
        self.ylim = ylim
        self.boundary = boundary
        self.normalize_with_neighbour_num = normalize_with_neighbour_num

        self.num_iterations = num_iterations
        self.bifurctaion_statistics = [
            ModelStatistics(
                self.num_iterations,
                int(N),
                (R_op, R_sp),
                alpha,
                beta,
                sigma_op,
                sigma_sp,
                delta_t,
                init_xlim,
                init_ylim,
                init_oplim,
                useDoublewell,
                spatial_clustering,
                opinion_clustering,
                dynamic_eps,
                useHDBSCAN,
                datasim,
                xlim,
                ylim,
                boundary,
                normalize_with_neighbour_num,
                input_df,
                initialize,
            )
            for N, R_op, R_sp, alpha, beta, sigma_op, sigma_sp, in self.parameters
        ]

        self.steps = self.bifurctaion_statistics[0].steps
        self.data_bank = data_bank
        self.load_from_file = True if self.data_bank is not None else False

    def simulate(self, steps, n_processes=-1, clear=False):
        for statistics in self.bifurctaion_statistics:
            if not statistics.data_exists(steps):
                print(f"There are {joblib.cpu_count()} processes available.")
                print(f"Simulation started.")
                statistics.simulate(steps, n_processes)
                if self.data_bank is not None:
                    print("Saving Data.")
                    statistics.save_results_as_numpy(self.data_bank)
                    if clear:
                        statistics.clear()
            else:
                print(f"Data Exits")
                # statistics.load_results_from_numpy(self.data_bank)
        self.steps = self.bifurctaion_statistics[0].steps

    def diversity_metrics(self, methods: list, timestep: Optional[int] = None):
        """
        This function serves as a general function that calls the implemented diversity metrics. Allowed values for
        methods are "average_local_agreement", "polarization", "opinion_diversity", "network_disagreement_index",
        "polarization_level" and "radicalisation_level".

        Parameters
        ----------
        methods : list of str
            diversity metrics to be calculated.
            methods must belong to the set {"average_local_agreement", "polarization",
             "opinion_diversity", "network_disagreement_index", "polarization_level", "radicalisation_level"}

        Returns
        -------
        res : list
            list of calculated averaged diversity metrics in input order.

        """
        self.load_from_file = True if self.data_bank is not None else False

        valid = ("global_assortativity",)
        res = []
        if timestep is not None:
            if len(methods) > 1 and methods[0] != "global_assortativity":
                raise NotImplementedError(
                    "Metrics other than assortativity can only be calculated for"
                    " the last time step."
                )
        else:
            timestep = self.steps

        for method in methods:
            if method not in valid:
                warnings.warn(r"{} is not a valid method".format(method))
            elif method == "global_assortativity":
                metric = np.zeros((self.num_parameters, self.num_iterations))
                for j, statistics in enumerate(self.bifurctaion_statistics):
                    if self.load_from_file:
                        statistics.steps = self.steps
                        try:
                            statistics.load_results_from_numpy(self.data_bank)
                            err = False
                        except ValueError as e:
                            err = True
                            print(e)
                    if not err:
                        metric[j, :] = statistics.global_assortativity(
                            time_step=timestep
                        )
                    else:
                        metric[j, :] = np.NAN * np.zeros(statistics.N)
                    statistics.clear()
                metric = (np.mean(metric, axis=1), np.std(metric, axis=1))
                res.append(metric)
        return np.array(res)

    def plot_diversity_metrics(
        self,
        bifurcation_parmeter: str,
        methods,
        timestep: Optional[int] = None,
        ax: plt.Axes | None = None,
        print_title=True,
        print_legend: bool = True,
        xlabel: str | None = None,
        ylabel: str | None = None,
        marker: list | list[str] | None = None,
        marker_colors: list[str] | None = None,
        confidence_level=None,
        std_error=False,
    ):
        assert bifurcation_parmeter in self.bifurcation_parameters.keys()
        if ax is None:
            plt.gca()
        metrics = self.diversity_metrics(methods, timestep=timestep)
        for i, method in enumerate(methods):
            metric_mean, metric_std = metrics[i]
            ax.plot(
                self.bifurcation_parameters[bifurcation_parmeter],
                metric_mean,
                label=method,
                color="k",
            )
            if std_error:
                ax.errorbar(
                    self.bifurcation_parameters[bifurcation_parmeter],
                    metric_mean,
                    yerr=metric_std,
                    fmt="None",
                    capsize=3,
                    color="k",
                )
            # plot the individual markers separately
            if marker is not None:
                if isinstance(marker, str):
                    marker = [
                        marker
                        for _ in self.bifurcation_parameters[bifurcation_parmeter]
                    ]
                if isinstance(marker_colors, str):
                    marker_colors = [
                        marker_colors
                        for _ in self.bifurcation_parameters[bifurcation_parmeter]
                    ]
                for p, parameter in enumerate(
                    self.bifurcation_parameters[bifurcation_parmeter]
                ):
                    ax.scatter(
                        parameter,
                        metric_mean[p],
                        marker=marker[p],
                        c=marker_colors[p],
                        zorder=10,
                    )
            # plot confidence intervalls
            if confidence_level is not None:
                raise NotImplementedError
        ax.grid(True)
        if print_legend:
            ax.legend()
        if xlabel is None:
            ax.set_xlabel(bifurcation_parmeter)
        else:
            ax.set_xlabel(xlabel)
        if ylabel is not None:
            ax.set_ylabel(ylabel)
        if print_title:
            plt.suptitle(
                r"Order Parameters with varied {}  after {} runs at time $t ={:.2f}$".format(
                    bifurcation_parmeter, self.num_iterations, self.steps * self.delta_t
                )
                + "\n"
                + r"for $N ={}, R_{{op}} ={}, R_{{sp}} ={}, \alpha ={}, \beta ={}, \sigma_{{op}} ={},"
                r" \sigma_{{sp}} ={}$".format(
                    self.parameters[0][0],
                    self.parameters[0][1],
                    self.parameters[0][2],
                    self.parameters[0][3],
                    self.parameters[0][4],
                    self.parameters[0][5],
                    self.parameters[0][6],
                )
            )
        return ax

    def plot_assortativity_time_series(
        self,
        axs: np.ndarray[plt.Axes] | plt.Axes,
        max_timestep: int | None = None,
        title=None,
        plot_all_realizations=False,
        plot_stdw=False,
        parameter_subset: list[int] | None = None,
    ) -> plt.Axes | np.ndarray[plt.Axes]:
        """
        Plot the mean  assortativity time series for a Parameter Sweep either in one or several axes
        objects. Possibly plot all the all realizations or the standard deviations envelope.
        @param axs: optinal axes object
        @param title: optinal, bool
            plot
        @param plot_all_realizations:
        @param plot_stdw:
        @return:
        """
        # Initial set up and checking inputs
        assert isinstance(axs, plt.Axes) or isinstance(axs, np.ndarray)
        if isinstance(axs, np.ndarray):
            assert all(isinstance(ax, plt.Axes) for ax in axs)
            if parameter_subset is None:
                parameter_subset = [
                    i for i, _ in enumerate(self.bifurctaion_statistics)
                ]
            else:
                assert len(parameter_subset) == len(axs)
        elif parameter_subset is not None:
            assert (
                len(parameter_subset) <= 3
            )  # Currently not enough colors implemented.
        else:
            assert len(self.bifurctaion_statistics) <= 3
            parameter_subset = [i for i, _ in enumerate(self.bifurctaion_statistics)]
        if parameter_subset is not None:
            assert all(isinstance(value, int) for value in parameter_subset)
        if max_timestep is None:
            max_timestep = self.steps + 1

        # main loop to plot the timeseries
        for j, statistics in enumerate(self.bifurctaion_statistics):
            if parameter_subset is not None:
                if j not in parameter_subset:
                    continue
            if self.load_from_file:
                statistics.load_results_from_numpy(self.data_bank)
            # Set the axes to which one parameter set is plotted to
            if isinstance(axs, np.ndarray):
                ax = axs[j]

                colors = ["tab:grey" for _ in range(self.num_parameters)]
            else:
                ax = axs
                # set the colormap for the different parameters
                colors = ["tab:purple", "tab:red", "tab:green"]
            # calculate the global assortativity
            assorts = np.zeros((self.num_iterations, max_timestep))
            for i in range(max_timestep):
                assorts[:, i] = statistics.global_assortativity(time_step=i)
            value = list(self.bifurcation_parameters.values())[0][j]
            # radius = self.bifurcation_parameters["R_{op}"][j]
            parameter_name = list(self.bifurcation_parameters.keys())[0]
            # plot the mean assortativity
            ax.plot(
                self.delta_t * np.arange(max_timestep),
                assorts.mean(axis=0),
                label=parameter_name + f"$= {value}$",
                color=colors[parameter_subset.index(j)],
            )

            # plot all the realizations in thin grey lines to illustrate stochastic behaviour
            if plot_all_realizations:
                for i, model in enumerate(statistics.models):
                    ax.plot(
                        statistics.delta_t * np.arange(statistics.steps + 1),
                        assorts[i],
                        color=colors[j],
                        lw=0.5,
                        alpha=0.4,
                    )
            statistics.clear()
            # Plot filled standard deviation around the mean
            if plot_stdw:
                ax.fill_between(
                    self.delta_t * np.arange(max_timestep),
                    assorts.mean(axis=0) - assorts.std(axis=0),
                    assorts.mean(axis=0) + assorts.std(axis=0),
                    color=colors[parameter_subset.index(j)],
                    alpha=0.2,
                )

            ax.legend()
            ax.set_ylim((-0.1, 1))
            ax.set_ylabel("Global Assortativity")
            ax.set_xlabel("Time t")
            if title is not None:
                ax.set_title(title)
            ax.grid(True)
        return axs


if __name__ == "__main__":
    # model parameters for a base model
    N = 100
    R_sp = 0.15
    alpha = 20
    beta = 20
    sigma_op = 0.05
    sigma_sp = 0.05
    delta_t = 0.02
    local_coupling = True
    KDE_bandwith = 0.02
    timestep = 25

    # simulation parameters
    n_iter = 8
    num_processes = 2
    # plotting_parameters -> For all figures
    deg_lim = (0, 0.26)
    opinion_ylim = (0, 0.2)
    xlim = (-0.5, 0.5)
    ylim = (-0.5, 0.5)

    bins = 100
    target_folder = "Images/Test"

    R_op = np.concatenate([np.linspace(0.03, 0.148, 10), np.linspace(0.152, 0.3, 10)])
    parameters = [N, R_op, R_sp, alpha, beta, sigma_op, sigma_sp]
    model = BifurcationAnalysis(
        100, parameters, delta_t, normalize_with_neighbour_num=local_coupling
    )
    model.simulate(500)
    model.plot_diversity_metrics(
        r"R_{op}",
        [
            "average_local_agreement",
            "polarization",
            "opinion_diversity",
            "network_disagreement_index",
            "polarization_level",
            "radicalisation_level",
        ],
    )
