from math import ceil
import numpy as np
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

def get_bins(data, opinion):
    """
    Compute the bin edges for creating a histogram of the dataset.
    The bin size ich chosen, such that each unique value is contained in a separate bin
    @param data: The pandas dataframe containing the preprocessed GSS data
    @param opinion: string, the key for accessing the dataframe
    @return: bin edges covering the entire data
    """
    d = np.diff(np.unique(data[opinion])).min()
    left_of_first_bin = data[opinion].min() - float(d) / 2
    right_of_last_bin = data[opinion].max() + float(d) / 2

    return np.arange(left_of_first_bin, right_of_last_bin + d, d)


def assign_gridcells_7x7(data):
    """
    Assigns grid cell indices to the data based on the values of two columns, "partyid" and "polviews".
    The grid is 7x7 in size. The Gridcells are stored in the dataframe provided to the function.
    @param data: The pandas dataframe containing the preprocessed GSS data
    @return: The dataframe containing the grid cell indices as well as the GSS Data
    """
    temp_party = np.around(data["partyid"], 5)
    temp_pol = np.around(data["polviews"], 5)
    party_values = sorted(temp_party.unique())
    pol_values = sorted(temp_pol.unique())

    grid_dict = {}
    # generate dictionary to quickly assign cell index from coordinates
    for n1 in range(len(party_values)):
        for n2 in range(len(pol_values)):
            grid_dict[n2 * len(party_values) + n1] = (party_values[n1], pol_values[n2])

    # pass over agents and assign cell index
    gridcell = []

    for j in zip(temp_party, temp_pol):
        gridcell.append(list(grid_dict.keys())[list(grid_dict.values()).index(j)])
    data["gridcell"] = gridcell

    return data


def get_grid_cell(x_min, x_max, y_min, y_max, coordinate):
    #   returns grid cell counting like this
    #   6 7 8
    #   3 4 5
    #   0 1 2
    x, y = coordinate

    if x < x_min or x > x_max or y < y_min or y > y_max:
        return None  # Coordinate outside the grid

    cell_width = (x_max - x_min) / 3  # Assuming 3 cells horizontally
    cell_height = (y_max - y_min) / 3  # Assuming 3 cells vertically

    cell_x = int((x - x_min) // cell_width)
    cell_y = int((y - y_min) // cell_height)  # Reverse the row enumeration

    return cell_x + cell_y * 3


def grid_uniform(data, seed=123):
    """
    A new and smaller grid is computed.
    Redistributes agents uniformly within each grid cell, and updates the grid cell indices based on
    the new agent positions.
    @param data:
    @param seed:
    @return:
    """
    rng = np.random.default_rng(seed)
    # ==============================================================================================
    # Is this not just the boarders nxn function?
    party_values = sorted(data["partyid"].unique())
    pol_values = sorted(data["polviews"].unique())

    # generate a dictionary that contains the x and y values of the borders around a grid cell
    diff_party = np.diff(party_values).min()
    diff_pol = np.diff(pol_values).min()

    border_party = [round(x - (diff_party / 2), 5) for x in party_values]
    border_party.append(round(party_values[-1] + (diff_party / 2), 5))

    border_pol = [round(x - (diff_pol / 2), 5) for x in pol_values]
    border_pol.append(round(pol_values[-1] + (diff_pol / 2), 5))

    borders = {}
    for n1 in range(len(party_values)):
        for n2 in range(len(pol_values)):
            borders[n1 * len(party_values) + n2] = (
                border_party[n2],
                border_party[n2 + 1],
                border_pol[n1],
                border_pol[n1 + 1],
            )
    # ==============================================================================================

    # finds the gridcells for all the responses in the dataframe and adds them to the dataframe
    data_with_gridcells = assign_gridcells_7x7(data)

    # look up borders for grid cell and uniformly redistribute agents within the 7x7 gridcells
    for j in data_with_gridcells.index:
        left, right, bottom, top = borders.get(data_with_gridcells.loc[j, "gridcell"])
        data_with_gridcells.loc[j, "partyid"] = rng.uniform(left, right)
        data_with_gridcells.loc[j, "polviews"] = rng.uniform(bottom, top)

    party_min, party_max = min(border_party), max(border_party)
    pol_min, pol_max = min(border_pol), max(border_pol)

    grid_cell_uniform = []

    temp_party_uniform = np.around(data["partyid"], 5)
    temp_pol_uniform = np.around(data["polviews"], 5)

    # assign new grid cell indexes based on the 3x3 grid
    for j in zip(temp_party_uniform, temp_pol_uniform):
        grid_cell_uniform.append(
            get_grid_cell(party_min, party_max, pol_min, pol_max, j)
        )

    data_with_gridcells["gridcell"] = grid_cell_uniform

    return data_with_gridcells


def borders_3x3_small(data):
    """
    Assigns grid cell indices to the data based on the values of two columns, "partyid" and "polviews".
    The Gridcells are stored in the dataframe provided to the function.
    @param data: The pandas dataframe containing the preprocessed GSS data
    @return: The dataframe containing the grid cell indices as well as the GSS Data
    """
    # decreases the size of the box in which agents are initialized
    min_party, max_party = min(data["partyid"]), max(data["partyid"])
    min_pol, max_pol = min(data["polviews"]), max(data["polviews"])

    border_party = np.round(
        np.linspace(min_party, max_party, 4), 5
    )  # splitting into 4 results in 3 cells/sections
    border_pol = np.round(np.linspace(min_pol, max_pol, 4), 5)

    diff = np.round(
        np.abs(border_party[0] - border_party[1]) / 4, 5
    )  # compute the offset of the borders -> move 1/4 of distance towards middle

    borders = {}
    for n1 in range(3):
        for n2 in range(3):
            borders[n1 * 3 + n2] = (
                border_party[n2] + diff,
                border_party[n2 + 1] - diff,
                border_pol[n1] + diff,
                border_pol[n1 + 1] - diff,
            )

    return borders


def get_histogram_counts(data, opinion, bins, method, density=False):
    """
    Calculates histogram counts for each grid cell based on the specified opinion column and method.
    @param data:
    @param opinion:
    @param bins:
    @param method:
    @param density:
    @return:
    """
    histogram_counts = []
    max_count = 0
    if method == "merge":
        grid3x3 = {
            0: [0, 1, 7, 8],
            1: [2, 3, 4, 9, 10, 11],
            2: [5, 6, 12, 13],
            3: [14, 15, 21, 22, 28, 29],
            4: [16, 17, 18, 23, 24, 25, 30, 31, 32],
            5: [19, 20, 26, 27, 33, 34],
            6: [35, 36, 42, 43],
            7: [37, 38, 39, 44, 45, 46],
            8: [40, 41, 47, 48],
        }
    for j in range(9):
        # isolate agents on a grid point
        if method == "uniform":
            temp = data[data["gridcell"] == j]
        if method == "merge":
            temp = data[data["gridcell"].isin(grid3x3[j])]
        counts, _ = np.histogram(temp[opinion], bins, density=density)
        histogram_counts.append(counts)
        # calculate maximum count for shared axis
        if max(counts) > max_count:
            max_count = max(counts)
    # round max to next 10
    max_count = int(ceil(max_count / 10.0)) * 10

    return histogram_counts, max_count
