"""
This file contains the GSS class handling all the functions that are related to the GSS dataset.
The function parameter search() is used to sample the parameter space.
"""

import os
from typing import Optional, List

import tqdm
import numpy as np
import pandas as pd
from scipy.stats import entropy
import matplotlib.pyplot as plt
from joblib import Parallel, delayed
from sklearn.preprocessing import MinMaxScaler
from pandas.core.groupby import DataFrameGroupBy

from src import Model
from src.gss import Data_simulation_util as Util


class GSS:
    """This class handles all interactions with the GSS dataset. Loading and preparing the relevant
    ata, as well as comparing simulations, etc."""

    # Type hinting all the attributes, that will be set by methods other than __init__
    gss_data: None | pd.DataFrame | pd.Series
    gss_data_groups: Optional[DataFrameGroupBy]
    number_of_agents: Optional[int]
    years: Optional[np.ndarray]
    time_period: Optional[int]
    xlim: Optional[tuple[float, float]]
    ylim: Optional[tuple[float, float]]
    bins: Optional[np.ndarray]
    initial_data: Optional[pd.DataFrame]
    model: Optional[Model]
    kl_divergence: Optional[List[int]]
    gss_histograms: Optional[List]
    simulation_histograms: Optional[List]
    distributed_gss_data: Optional[dict]

    def __init__(
        self, data_key: str, first_year: Optional[int] = None, last_year: int = 2018
    ) -> None:
        """
        Initializes the GSS class with the specified data key and optional year range.

        Parameters:
            data_key (str):
                Key to identify the dataset. Must be either 'helpsick' or 'eqwlth'.
            first_year (Optional[int]):
                The starting year for the data comparison. If not specified,
                                        defaults to 1975 for 'helpsick' and 1978 for 'eqwlth'.
            last_year (int):
                The final year for the data comparison. Defaults to 2018.

        Attributes set in __init__:
            directory (str):
                Directory path for storing images related to GSS data analysis.
            data_key (str):
                Key to identify the dataset.
            first_year (int):
                The starting year for the data comparison.
            last_year (int):
                The final year for the data comparison.
            gss_data (Optional[pd.DataFrame]):
                Initially None, will contain the raw GSS dataset.
            gss_data_groups (Optional[DataFrameGroupBy]):
                Initially None, will contain groups of answers for each year.
            number_of_agents (Optional[int]):
                Initially None, will hold the number of agents in the simulation.
            gss_participants (Optional[int]):
                Initially None, will hold the number of participants in the GSS dataset
                                              for each year.
            years (Optional[np.ndarray]):
                Initially None, will hold the list of years for which data is available.
            time_period (Optional[int]):
                Initially None, will hold the length of the investigated time period in years.
            xlim (Optional[tuple[float, float]]):
                Initially None, boundary condition for the simulation.
            ylim (Optional[tuple[float, float]]):
                Initially None, boundary condition for the simulation.
            bins (Optional[np.ndarray]):
                Initially None, will hold the bins for histogram analysis.
            initial_data (Optional[pd.DataFrame]):
                Initially None, will hold the initial data of the simulation.
            model (Optional[Model]):
                Initially None, will hold an instance of the Model class handling the simulation.
            kl_divergence (Optional[List[int]]):
                Initially None, will hold the time series of the KL divergence.
            gss_histograms (Optional[List[int, np.ndarray]]):
                Initially None, will hold the histograms of GSS data and the largest
                count.
            simulation_histograms (Optional[List[int, np.ndarray]]):
                Initially None, will hold the histograms of simulation data and the largest
                count.
            distributed_gss_data (Optional[dict[int, pd.DataFrame]]):
                Initially None, will hold the distributedanswers of the GSS dataset
                for each year.
        """
        assert data_key in ["helpsick", "eqwlth"]
        self.data_key = data_key  # pandas key for the relevant dataset
        self.first_year = first_year  # the year from which the comparison is started
        if self.first_year is None:
            self.first_year = 1975 if self.data_key == "helpsick" else 1978
        self.last_year = last_year  # final year of the data comparison
        self.gss_data = None  # Contains the raw data set
        self.gss_data_groups = None  # Contains groups of answers for each year
        self.number_of_agents = None  # The number of agents in the simulation
        self.gss_participants = (
            None  # List of the number of particiapnts in the gss dataset
        )
        # for each year
        self.years = None  # List of the year for which we have data available
        self.time_period = None  # Length of the investigated time period in years
        self.xlim = None  # boundary condition for the simulation
        self.ylim = None  # boundary condition for the simulation
        self.bins = None
        self.initial_data = (
            None  # Dataframe containing the initial data of the simulation
        )
        self.model = None  # Instance of the Model class handling the simulation
        self.kl_divergence = None  # time series of the kl divergence
        self.gss_histograms = None
        self.simulation_histograms = None
        self.distributed_gss_data = (
            None  # List of dataframes containing the distributed answers of
        )

        # the GSS data set for each year

    def data_preparation(self, seed: Optional[int] = 123) -> None:
        """
        Processes the raw data files and prepares histograms for comparison with simulation results.
        Generates initial data for the simulation.

        Parameters:
            seed (Optional[int]): Defines the seed for the spreading of the data. If None is provided,
                                  the random number generator will use None.

        This method performs the following steps:
            1. Reads only the specified columns ('year', 'partyid', 'polviews', and the column
            specified by self.data_key).
            2. Filters the data to include only usable values within the desired time period.
            3. Rescales the data using the rescale_data method.
            4. Determines the unique years for which data is available and filters them based on the
            specified time range.
            5. Groups the data by year for easy access.
            6. Sets the input dataframe containing the initial data for the simulation.
            7. Computes the bin edges for histogram analysis.
            8. Determines the number of agents and the length of the investigated time period.
            9. Disperses the agents in the social space and assigns them to a 3x3 grid.
           10. Computes the histograms for the initial data.
           11. Sets the boundary conditions for the social space.

        Returns:
            None
        """

        # Construct the absolute path to the GSS.xlsx file
        base_dir = os.path.dirname(__file__)
        file_path = os.path.join(
            base_dir, "..", "..", "data", "GSS", "GSS.xlsx"
        )

        # Specify the relevant columns
        columns_to_read = ("year", "partyid", "polviews", self.data_key)

        # Read only the specified columns
        data = pd.read_excel(file_path, usecols=columns_to_read)

        if self.last_year is None:
            self.last_year = data["year"].max()
        # get usable values for the desired timeperiod the data set contains unusable data point,
        # e.g., people answering only a subset of the question etc.
        self.gss_data = data[
            (data["partyid"].isin(np.arange(0, 7)))
            & (data["polviews"].isin(np.arange(1, 8)))
            & (data[self.data_key].isin(np.arange(1, 6)))
            & (data["year"] <= self.last_year)
        ]

        types_dict = {
            "year": int,
            "partyid": float,
            "polviews": float,
            self.data_key: str,
        }
        self.gss_data = self.gss_data.astype(types_dict)
        self.rescale_data()

        # get the years for which we have data
        self.years = np.array(sorted(self.gss_data["year"].unique()))
        self.years = np.array(
            [year for year in self.years if self.first_year <= year <= self.last_year]
        )

        # allows us to easily access all the data groups for a given year
        self.gss_data_groups = self.gss_data.groupby(self.gss_data["year"])

        # set the input dataframe containing the initial data for the simulation
        first_year_data = self.gss_data_groups.get_group(self.first_year)

        # get the bin edges, where the width is the difference between the two smallest unique
        # values in the data set
        self.bins = Util.get_bins(self.gss_data, opinion=self.data_key)

        self.number_of_agents = len(first_year_data)
        self.time_period = max(self.years) - self.first_year

        # disperse the agents in the social space and assigns them to a 3x3 grid
        temp_initial_data = Util.grid_uniform(first_year_data.copy(), seed=seed)

        # Assign agents to their respective gridcells
        Util.borders_3x3_small(first_year_data)

        # get the histograms for the initial data
        init_counts, _ = Util.get_histogram_counts(
            temp_initial_data, opinion=self.data_key, bins=self.bins, method="uniform"
        )

        # for compatibility with the Model class remove gridcells from the dataframe
        self.initial_data = temp_initial_data[
            ["partyid", "polviews", self.data_key]
        ].copy()

        # setting the limits for the social space, that will be enforced as boundary conditions
        self.xlim = (
            1.1 * min(self.initial_data["partyid"]),
            1.1 * max(self.initial_data["partyid"]),
        )
        self.ylim = (
            1.1 * min(self.initial_data["polviews"]),
            1.1 * max(self.initial_data["polviews"]),
        )

    def data_analysis(
        self,
        R: tuple[float, float],
        alpha: float | int,
        beta: float | int,
        sigma_op: float | int,
        sigma_sp: float | int,
        seed: int | None = 123,
        model: Model | None = None,
        boundary: bool = True,
    ) -> None:
        """
        Simulates a realization using the initial data provided by the data_preparation function,
        computes KL divergences, and generates histograms for both the GSS data and simulation results.

        Parameters:
            R (tuple[float, float]):
                Radii parameter for spatial interaction.
            alpha (float | int):
                Opinion interaction strength.
            beta (float | int):
                Spatial interaction strength.
            sigma_op (float | int):
                Opinion noise intensity.
            sigma_sp (float | int):
                Spatial noise intensity.
            seed (int):
                The seed used for the random number generator in the simulation.
            model (Model | None):
                If the simulation was computed elsewhere, an instance of the Model
                                  class can be passed here.
            boundary (bool):
                Whether to enforce boundary conditions during the simulation.

        Returns:
            None

        Raises:
            ValueError:
            If the initial data or number of agents is not set by running data_preparation() first.

        This method performs the following steps:
            1. Initializes the Model class with the given parameters if no model is provided.
            2. Runs the simulation for the specified time period.
            4. For each time step, generates histograms for the GSS data and simulation results.
            5. Computes KL divergences between the GSS data and simulation results for each grid
            cell and the overall distribution.
            6. Stores histogram data and KL divergence values for later analysis.
        """
        if self.initial_data is None or self.number_of_agents is None:
            raise ValueError("Run data_analysis() first.")
        # Initialize and simulate the model
        if model is None:
            self.model = Model(
                self.number_of_agents,
                R,
                alpha,
                beta,
                sigma_op,
                sigma_sp,
                0.01,
                datasim=True,
                xlim=np.round(self.xlim, 3),
                ylim=np.round(self.ylim, 3),
                boundary=boundary,
                seed=seed,
            )

            self.model.initialize_model(self.initial_data)

            # start the simulation with a step corresponding to one year
            self.model.simulate(self.time_period + 1)

        # Start of the data analysis
        simulation_data = self.model.get_results().copy()
        self.gss_histograms = []
        self.simulation_histograms = []
        self.kl_divergence: list[float] = []
        self.gss_participants = []
        self.distributed_gss_data = {}

        # generate histograms and compare gss data to simulation results
        for time_step in range(self.time_period + 1):
            # We only need the simulation results if we have data for those years
            if time_step + self.first_year in self.gss_data_groups.groups.keys():
                # we need the right format to pass the simulation data to utils.get_histograms()
                temp_simulation_data = simulation_data["Coordinates"][time_step].copy()
                temp_simulation_data["opinions"] = list(
                    simulation_data["Opinions"][time_step].copy().values
                )
                temp_simulation_data["gridcell"] = list(
                    simulation_data["gridcell"][time_step].copy().values
                )

                # generates histograms for the simulation results
                simulation_histograms = Util.get_histogram_counts(
                    temp_simulation_data,
                    opinion="opinions",
                    bins=self.bins,
                    method="uniform",
                )
                # get the relevant gss data
                gss_answers = self.gss_data_groups.get_group(
                    self.first_year + time_step
                )
                # Setting the participant numbers
                self.gss_participants.append(len(gss_answers))
                # disperse the raw survey data
                distributed_gss_answers = Util.grid_uniform(
                    gss_answers.copy(), seed=seed
                )

                # get histograms for the gss data
                gss_answer_histogram = Util.get_histogram_counts(
                    distributed_gss_answers,
                    opinion=self.data_key,
                    bins=self.bins,
                    method="uniform",
                )
                # Add the gss data to the class for later plotting of different time steps
                self.distributed_gss_data[time_step + self.first_year] = (
                    distributed_gss_answers[
                        ["partyid", "polviews", self.data_key]
                    ].copy()
                )
                # store the calculated histogram data
                self.gss_histograms.append(gss_answer_histogram)
                self.simulation_histograms.append(simulation_histograms)

                # Compute the KL divergence for the entire distribution
                self.kl_divergence.append(
                    self.compute_kl_divergence(
                        gss_answer_histogram[0], simulation_histograms[0]
                    )
                )

    def rescale_data(self):
        """
        Rescales the 'partyid' and 'polviews' columns of the GSS dataset to the range (-0.25, 0.25)
        and the column specified by 'data_key' to the range (-1, 1).
        This method uses MinMaxScaler from scikit-learn to normalize the data.  The 'partyid' and
        'polviews' columns are scaled to a smaller range suitable for spatial data, while the
        column specified by 'data_key' is scaled to a larger range suitable for opinion data.
        Attributes:
            gss_data (pd.DataFrame):
                The GSS dataset containing the columns to be rescaled.
            data_key (str):
                The key for the specific column in the GSS dataset to be rescaled.
        Returns:
            None
        """
        # defining a scaler that maps the data to the desired range
        scaler_spatial = MinMaxScaler((-0.25, 0.25))
        scaler_opinion = MinMaxScaler((-1, 1))

        # apply the scaler to the dataset
        self.gss_data.loc[:, ["partyid", "polviews"]] = scaler_spatial.fit_transform(
            self.gss_data[["partyid", "polviews"]]
        )
        self.gss_data.loc[:, self.data_key] = scaler_opinion.fit_transform(
            self.gss_data[self.data_key].to_numpy().reshape(-1, 1)
        )

    def compute_kl_divergence(self, data_counts, sim_counts):
        """
        Calculates the KL Divergence KL(data | simulation) between the joint distribution over
        social space and opinion space.

        Parameters:
            data_counts (list): Histogram counts for the joint distribution from the GSS data.
                Each entry in this list must contain the histogram counts for the opinion within a
                grid cell.Designed to take input from input,_ = Util.get_histogram_counts()
            sim_counts (list): Histogram counts for the joint distribution from the simulated data.
                Each entry in this list must contain the histogram counts for the opinion within a
                grid cell.Designed to take input from input,_ = Util.get_histogram_counts()

        Returns:
            float: The KL divergence value between the GSS data and the simulated data.

        The method  applies pseudo counts to avoid division by zero.
        """

        n_data = sum(sum(data_counts))
        n_sim = sum(sum(sim_counts))
        data_counts = (np.concatenate(data_counts) + np.ones(45)) / (n_data + 45)
        sim_counts = (np.concatenate(sim_counts) + np.ones(45)) / (n_sim + 45)
        kl_divergence = entropy(data_counts, sim_counts)
        return kl_divergence

    def goodness_of_fit(self) -> float:
        """
        A simple heuristic to evaluate the goodness of fit by integrating the KL divergence against
        an exponential decay, to prioritize the fit for times closer to the initialization.

        @return: float
            The goodness of fit.
        """
        if self.kl_divergence is None:
            raise ValueError(
                "KL divergence is not defined. Run self.data_analysis() first."
            )
        envelope = np.exp(-0.1 * self.years)
        return np.trapz(y=envelope * self.kl_divergence, x=self.years) * 10**87
        # Multiplication with a constant does not change ordering

    # =============================================================================
    # Plotting
    # =============================================================================

    def plot_kl_divergence(
        self, filename: str | None = None, ax=None
    ) -> Optional[plt.Axes]:
        """ "
        Plots the KL divergence timeseries.

        Parameters:
            filename (str, optional):
            If specified, the plot will be saved to this file. The file extension has to be
            included. Default is None.
            ax (matplotlib.axes.Axes, optional):
            he timeseries will be plotted on this axis. Default is None. If None is given, a new
            plot will be created.

        Returns:
            Optional[plt.Axes]: Matplotlib Axes object containing the KL divergences if no filename
            is provided.

        Raises:
            ValueError: If kl_divergence is not defined, indicating that data_analysis() needs to be
            run first.
        """
        if self.kl_divergence is None:
            raise ValueError("kl_divergence is not defined. Run data_analysis() first.")
        if ax is None:
            fig, ax = plt.subplots(figsize=(6, 3))
        plt.tight_layout()
        ax.plot(self.years, self.kl_divergence, marker="d")
        ax.grid(True)
        ax.set_axisbelow(True)
        if filename is not None:
            plt.savefig(filename)
            plt.close()
        else:
            return ax

    def plot_distributed_data(
        self,
        year,
        radii,
        xlim,
        ylim,
        boundary=True,
        use_R_sp=True,
        filename=None,
        figure=None,
        s=None,
        network_alpha=None,
        colorbar=True,
        network_lw=None,
        hist_ylim=None,
        node_linewidth=0,
    ):

        model = Model(
            len(self.distributed_gss_data[year]),
            radii,
            1,
            1,
            1,
            1,
            0.01,
            datasim=True,
            xlim=np.round(xlim, 3),
            ylim=np.round(ylim, 3),
            boundary=boundary,
        )
        model.initialize_model(self.distributed_gss_data[year])
        fig = model.plot_with_marginals(
            xlim=xlim,
            ylim=ylim,
            print_title=False,
            xlabel=r"democrat $\longleftrightarrow$ republican",
            ylabel=r"liberal $\longleftrightarrow$ conservative",
            use_spatial_network=use_R_sp,
            filename=filename,
            bins=8,
            opinion_label=r"Government should reduce differences $\longleftrightarrow$ No "
            r"government action",
            no_ticks=True,
            fig=figure,
            s=s,
            network_alpha=network_alpha,
            colorbar=colorbar,
            network_lw=network_lw,
            hist_ylim=hist_ylim,
            node_linewidth=node_linewidth,
        )
        return fig

    def plot_gss_participants(self, ax: Optional[plt.Axes] = None) -> plt.axes:
        data_set_name = {"eqwlth": "Equal Wealth", "helpsick": "Help Sick"}
        if self.gss_participants is None:
            raise LookupError(
                "You have to run data_analysis() before executing plot_gss_participants()"
            )
        if ax is None:
            ax = plt.gca()
        ax.plot(
            self.years,
            self.gss_participants,
            marker="d",
            label=data_set_name[self.data_key],
        )
        ax.legend()
        ax.grid()
        ax.set_xlabel("Year")
        ax.set_ylabel("Number of Participants")
        return ax


def parameter_search(
    data_key: str,
    sample_size=100,
    seed: Optional[int] = 123,
    first_year=None,
    last_year=None,
) -> None:
    """
    This function performs a uniform sampling from the parameter space and analyzes the KL
    divergence between the opinion data gathered from the GSS data set and the simulated
    distribution. The timeseries of the KL divergence will be plotted and saved. The filename
    contains the used parameter values and the "goodness-of-fit".

    Parameters:
        data_key (str):
            Key for the relevant GSS dataset ('helpsick' or 'eqwlth').
        sample_size (int, optional):
            Number of parameter sets to sample. Default is 100.
        seed (Optional[int], optional):
            Seed for the random number generator. Default is 123.
        first_year (Optional[int], optional):
            First year for the data comparison. Default is None.
        last_year (Optional[int], optional):
            Last year for the data comparison. Default is None.

    Returns:
        None
    """
    # initialization and preparation of the data
    gss_data = GSS(data_key=data_key, first_year=first_year, last_year=last_year)
    gss_data.data_preparation(seed=seed)

    # freeing up some memory, because a bunch of copies will be created
    # gss_data.gss_data = None

    # this function will be parallelized by joblib.Parralel
    def sample_parameters(gss_data_set, seed, data_key):
        temp_gss_data = gss_data_set.copy()
        alpha = np.random.uniform(low=1, high=50)
        beta = np.random.uniform(low=1, high=50)
        if data_key == "helpsick":
            parameters = [(0.15, 0.15), alpha, beta, 0.05, 0.05]
        else:
            R = np.random.uniform(low=0.07, high=0.3)
            parameters = [(R, R), alpha, beta, 0.1, 0.1]
        temp_gss_data.data_analysis(*parameters, seed=seed)

        # plot the kl divergence for each run, so we don't entirely rely on our heuristic
        # measure of how well data and simulation agree
        fig, ax = plt.subplots()
        temp_gss_data.plot_kl_divergence(ax=ax)
        base_dir = os.path.dirname(__file__)
        file_path = os.path.join(
            base_dir,
            f"../../data/DataBank/GSS/full_KL_help_sick_GF"
            f"={np.round(temp_gss_data.goodness_of_fit(), 3)}"
            f"_alpha={np.round(alpha, 2)}, "
            f"beta= {np.round(beta, 2)}.png",
        )

        plt.savefig(file_path)

        result = np.array(
            [
                temp_gss_data.goodness_of_fit(),
                parameters[0][0],
                parameters[0][1],
                alpha,
                beta,
                parameters[3],
                parameters[4],
            ]
        )
        return np.round(result, decimals=4)

    # parralel sampling
    results = Parallel(n_jobs=-1)(
        delayed(sample_parameters)(gss_data, seed, gss_data.data_key)
        for _ in tqdm.trange(sample_size)
    )
    # Construct the absolute path to the *_parameter_search.csv files
    base_dir = os.path.dirname(__file__)
    file_path = os.path.join(
        base_dir, f"../../data/DataBank/GSS/{data_key}_parameter_search.csv"
    )
    # storing the results
    data_base = pd.read_csv(file_path)
    for row in results:
        data_base.loc[len(data_base)] = row
    data_base.sort_values(by=["Goodness_of_Fit"], inplace=True)
    data_base.to_csv(file_path, index=False)


if __name__ == "__main__":
    parameters = ((0.08, 0.08), 2.4, 15.4, 0.1, 0.1)
    election_years = (1978, 2000, 2016)
    equal_wealth = GSS("eqwlth")
    equal_wealth.data_preparation()
    equal_wealth.data_analysis(*parameters)
