"""
Image Combination Module

This module provides functions to combine multiple images in various ways.
You can stack images vertically, place them side by side, or combine multiple
PNG images into a single image.

Functions:
----------
1. combine_images_side_by_side(image1_path: str, image2_path: str, output_path: str) -> None:
    Combine two images side by side and save the resulting image.

2. stack_images_vertically(image_paths: List[str], output_path: str) -> None:
    Stack multiple images vertically and save the resulting image.

3. combine_pngs(png1_path: str, png2_path: str, png3_path: str, png4_path: str, png5_path: str,
    png6_path: str, output_path: str) -> None:
    Combine six PNG images into a single image and save the resulting image.
"""

from PIL import Image


def combine_images_side_by_side(
    image1_path: str, image2_path: str, output_path: str
) -> None:
    """
    Combine two images side by side and save the resulting image.

    Parameters:
    -----------
    image1_path : str
        The file path to the first image.
    image2_path : str
        The file path to the second image.
    output_path : str
        The file path where the combined image will be saved.

    Returns:
    --------
    None

    Notes:
    ------
    - The function opens the two images specified in `image1_path` and `image2_path`.
    - It calculates the width and height of the resulting combined image.
    - A new blank image is created with the calculated dimensions.
    - The first image is pasted onto the new image at the leftmost position.
    - The second image is pasted onto the new image at the rightmost position.
    - The combined image is saved to `output_path`.

    Example usage:
    --------------
    combine_images_side_by_side("image1.png", "image2.png", "output.png")
    """
    # Open the images
    image1 = Image.open(image1_path)
    image2 = Image.open(image2_path)

    # Calculate the width and height of the combined image
    new_width = image1.width + image2.width
    new_height = max(image1.height, image2.height)

    # Create a new image with the calculated size
    combined_image = Image.new("RGBA", (new_width, new_height))

    # Paste the first image onto the new image at the leftmost position
    combined_image.paste(image1, (0, 0))

    # Paste the second image onto the new image at the rightmost position
    combined_image.paste(image2, (image1.width, 0))

    # Save the combined image
    combined_image.save(output_path)


def stack_images_vertically(image_paths, output_path):
    """
    Stack multiple images vertically and save the resulting image.

    Parameters:
    -----------
    image_paths : List[str]
        A list of file paths to the images to be stacked.
    output_path : str
        The file path where the combined image will be saved.

    Returns:
    --------
    None

    Notes:
    ------
    - The function opens all images specified in `image_paths`.
    - It calculates the width and height of the resulting combined image.
    - A new blank image is created with the calculated dimensions.
    - Each image is pasted onto the new image, stacked vertically.
    - The combined image is saved to `output_path`.

    Example usage:
    --------------
    stack_images_vertically(["image1.png", "image2.png", "image3.png"], "output.png")
    """
    # Open all images in the list
    images = [Image.open(image_path) for image_path in image_paths]

    # Calculate the width and height of the combined image
    max_width = max(image.width for image in images)
    total_height = sum(image.height for image in images)

    # Create a new image with the calculated size
    combined_image = Image.new(
        "RGBA", (max_width, total_height), color=(255, 255, 255, 0)
    )

    # Paste each image onto the new image, stacking them vertically
    y_offset = 0
    for image in images:
        combined_image.paste(image, (0, y_offset))
        y_offset += image.height

    # Save the combined image
    combined_image.save(output_path)


def combine_pngs(
    png1_path, png2_path, png3_path, png4_path, png5_path, png6_path, output_path
):
    """
    Combine six PNG images into a single image and save the resulting image.

    Parameters:
    -----------
    png1_path : str
        The file path to the first PNG image.
    png2_path : str
        The file path to the second PNG image.
    png3_path : str
        The file path to the third PNG image.
    png4_path : str
        The file path to the fourth PNG image.
    png5_path : str
        The file path to the fifth PNG image.
    png6_path : str
        The file path to the sixth PNG image.
    output_path : str
        The file path where the combined image will be saved.

    Returns:
    --------
    None

    Notes:
    ------
    - The function opens the six PNG images specified by the file paths.
    - It calculates the width and height of the resulting combined image.
    - A new blank image is created with the calculated dimensions.
    - Each PNG image is pasted onto the new image at the appropriate position.
    - The combined image is saved to `output_path`.

    Example usage:
    --------------
    combine_pngs("png1.png", "png2.png", "png3.png", "png4.png", "png5.png", "png6.png", "output.png")
    """
    # Open PNG images
    png1 = Image.open(png1_path)
    png2 = Image.open(png2_path)
    png3 = Image.open(png3_path)
    png4 = Image.open(png4_path)
    png5 = Image.open(png5_path)
    png6 = Image.open(png6_path)

    # Calculate dimensions of the combined image
    total_width = max(
        png1.width + png2.width + png3.width, png4.width + png5.width + png6.width
    )
    total_height = max(
        png1.height + png4.height, png2.height + png5.height, png3.height + png6.height
    )

    # Create a new blank image with the calculated size
    combined_image = Image.new("RGBA", (total_width, total_height), (255, 255, 255, 0))

    # Paste each image onto the combined image at the appropriate position
    combined_image.paste(png1, (0, 0))
    combined_image.paste(png2, (png1.width, 0))
    combined_image.paste(png3, (png1.width + png2.width, 0))
    combined_image.paste(png4, (0, png1.height))
    combined_image.paste(png5, (png1.width, png2.height))
    combined_image.paste(png6, (png1.width + png2.width, png3.height))

    # Save the combined image
    combined_image.save(output_path)
