from typing import Optional, Union

import numpy as np
import pandas as pd
from scipy.spatial.distance import cdist
from sklearn.cluster import DBSCAN
from sklearn.neighbors import NearestNeighbors
from kneed import KneeLocator
import hdbscan
import matplotlib.pyplot as plt

plt.rcParams["text.usetex"] = True

from src import Agent
from src.model_plot_lib import ModelPlotter


class Model:
    """
    Agent-based model using the class Agent. This class handles all calculations analysis and visualization for a single
    realization.
    """

    def __init__(
        self,
        N: int,
        R: float | tuple[float, float],
        alpha: float,
        beta: float,
        sigma_op: float,
        sigma_sp: float,
        delta_t: float,
        init_xlim: tuple[float, float] = (-0.25, 0.25),
        init_ylim: tuple[float, float] = (-0.25, 0.25),
        init_oplim: tuple[float, float] = (-1, 1),
        useDoublewell: bool = False,
        seed: Optional[int] = None,
        spatial_clustering: bool = True,
        opinion_clustering: bool = True,
        dynamic_eps: bool = False,
        useHDBSCAN: bool = True,
        datasim: bool = False,
        xlim: Optional[tuple[float, float]] = None,
        ylim: Optional[tuple[float, float]] = None,
        boundary: bool = True,
        normalize_with_neighbour_num: bool = False,
    ) -> None:
        """
        Constructor of the model that initializes class variables

        Parameters
        ----------
        N : int
            Number of agents in the system.
        R : float OR tuple of floats
            minimum radius to facilitate interaction between agents OR tuple defining radius for spatial and opinion
            interaction (opinion radius, spatial radius).
        alpha : float
            opinion strength.
        beta : float
            spatial strength.
        sigma_op : float
            opinion dynamics diffusion coefficient.
        sigma_sp : float
            spatial dynamics diffusion coefficient.
        delta_t : float
            time between simulation snapshots/steps.
        init_xlim : tuple of floats, optional
            defines lower and upper initial x bound (lower bound, upper bound). The default is (-0.25,0.25).
        init_ylim : tuple of floats, optional
            defines lower and upper initial y bound (lower bound, upper bound). The default is (-0.25,0.25).
        init_oplim : tuple of floats, optional
            defines lower and upper initial opinion bound (lower bound, upper bound). The default is (-1,1).
        useDoublewell : bool, optional
            specifying if the double well should be used as a potential landscape. The default is False.
        seed : int, optional
            seed for generation of random numbers. The default is None.
        spatial_clustering : bool, optional
            specifying if the clustering should include spatial coordinates. The default is True.
        opinion_clustering : bool, optional
            specifying if the clustering should include opinion coordinates. The default is True.
        dynamic_eps : bool, optional
            specifying if eps should be dynamically calculated before each clustering. The default is False.
        useHDBSCAN : bool, optional
            specifying if HDBSCAN should be used over DBSCAN. The default is True.
        datasim : bool, optional
            specifying if data is simulated (adds additional functionalities). The default is False.
        xlim : tuple of floats, optional
            defines lower and upper x bounds of the grid (lower bound, upper bound). The default is None.
        ylim : tuple of floats, optional
            defines lower and upper y bounds of the grid (lower bound, upper bound). The default is None.
        boundary : bool, optional
            specifying if agents are bounded within their grid cell. The default is True.
        normalize_with_neighbour_num: bool, optional
            if True, the interaction terms will be normalized by 1/deg(i) instead of 1/N. The default is False.

        Returns
        -------
        None.
        """

        self.N = N  # number if agents in the model
        self.agents = [Agent() for i in range(N)]  # initialize an agent list
        self.xdim = init_xlim  # interval for random initialization of x coordinates
        self.ydim = init_ylim  # interval for random initialization y coordinates
        self.opdim = init_oplim  # interval for random initialization of opinions
        if isinstance(
            R, tuple
        ):  # allow tuple input (R_op, R_sp) to define separate values, as well as integer input
            # which assumes R_sp = R_op
            self.R_op = R[0]
            self.R_sp = R[1]
        else:
            self.R_op = R  # opinion interaction radius
            self.R_sp = R  # spatial interaction radius
        self.alpha = alpha  # opinion strength
        self.beta = beta  # spatial strength
        self.sigma_op = (
            sigma_op  # opinion dynamics diffusion coefficient / opinion noise
        )
        self.sigma_sp = (
            sigma_sp  # spatial dynamics diffusion coefficient / spatial noise
        )
        self.delta_t = delta_t  # time between time steps
        self.distancematrix = np.zeros((self.N, self.N))  # initialize distance matrix
        self.rng = np.random.default_rng(
            seed
        )  # random number generator object for reproducible results
        self.seed = seed
        self.steps = 0  # internal counter of number of time steps
        self.normalize_with_neighbour_num = normalize_with_neighbour_num

        # parameters and variables for clustering
        self.useHDBSCAN = useHDBSCAN
        self.min_cluster_size = 5  # minimum cluster size (HDBSCAN)
        self.labels = [None] * N
        self.eps = (
            0.04  # minimum distance between agents in a cluster (DBSCAN and HDBSCAN)
        )
        self.dynamic_eps = dynamic_eps
        self.minpoints = 4  # minimum number of points to define a dense region (DBSCAN)
        self.spatial_clustering = (
            spatial_clustering  # include spatial coordinates for clustering
        )
        self.opinion_clustering = (
            opinion_clustering  # include opinio coordinates for clustering
        )
        self.clustering = spatial_clustering or opinion_clustering
        self.n_clusters = 0  # tracks the number of clusters
        self.hasNoise = (
            False  # checks if any agents are assigned to noise during clustering
        )

        # parameters for potential landscape
        self.useDoublewell = useDoublewell

        # Support for Data simulation
        self.datasim = datasim
        if self.datasim:
            self.ngrid = 3  # n x n dimension of the grid
            self.xlim = xlim  # x-axis bounds of the n x n grid
            self.ylim = ylim  # y-axis bounds of the n x n grid
            self.borders = {}  # dictionary containing the grid cell borders
            self.bound = boundary  # boolean: should boundary conditions be used
        self.gridcell = [None] * N  # grid cell membership of agents

        # Dictionary containing the model output for all iterations of the simulation
        # for more information see get_results()
        self.results = {}

    def initialize_model(self, input_df: Optional[pd.DataFrame] = None):
        """
        initialize the location and opinion of all agents randomly and calculate the distances and clustering.
         A input matrix can be provided to define the intitialization of the agents

        Parameters
        ----------
        input_df : Nx3 pd.DataFrame, optional
            Matrix containing x coordinates for all agents in the first column, y coordinates in the second column and
            opinions in the third column. The default is None.

        Returns
        -------
        None.

        """
        # reset model
        self.results = {}
        self.steps = 0

        # initialize agents
        if input_df is None:
            for i in range(self.N):  # uniform random initialization over the domain
                self.agents[i].set_x(self.rng.uniform(self.xdim[0], self.xdim[1]))
                self.agents[i].set_y(self.rng.uniform(self.ydim[0], self.ydim[1]))
                self.agents[i].set_opinion(
                    self.rng.uniform(self.opdim[0], self.opdim[1])
                )
        else:
            if input_df.shape != (self.N, 3):
                raise ValueError("Input dataframe does not match the initialized model")
            for i in range(self.N):  # initialize from input matrix
                self.agents[i].set_x(input_df.iloc[i, 0])
                self.agents[i].set_y(input_df.iloc[i, 1])
                self.agents[i].set_opinion(input_df.iloc[i, 2])

        # calculate distances and clustering
        self.calculate_distances()
        if self.clustering:
            self.cluster()
        if self.datasim:
            self.compute_borders()
            self.gridcell = [
                self.get_grid_cell(coord) for coord in self.get_coordinates()
            ]

        # initialize results dictionary
        self.results["Coordinates"] = [pd.DataFrame(self.get_coordinates())]
        self.results["Opinions"] = pd.DataFrame(self.get_opinions())
        self.results["Networks"] = [self.get_adjacency_matrix()]
        self.results["Cluster Labels"] = pd.DataFrame(self.labels)
        self.results["gridcell"] = pd.DataFrame(self.gridcell)

    def simulate(self, number_of_steps=1):
        """
        Simulate the model for a single step or multiple steps using the euler-maruyama method

        Parameters
        ----------
        number_of_steps : int, optional
            number of consecutive steps to be simulated . The default is 1.

        Returns
        -------
        None.

        """

        # initialize lists containing information on simulation snapshots
        coords = [None] * number_of_steps
        opinions = [None] * number_of_steps
        labels = [None] * number_of_steps
        gridcell = [None] * number_of_steps
        networks = [None] * number_of_steps

        for step in range(number_of_steps):
            new_x = np.zeros(self.N)
            new_y = np.zeros(self.N)
            new_op = np.zeros(self.N)
            for i in range(self.N):
                # get current data_is_saved
                x, y = self.agents[i].get_coordinates()
                op = self.agents[i].get_opinion()
                delta_x, delta_y = self.spatial_interaction(i)  # = U_tilde

                # euler maruyama
                x_new = (
                    x + delta_x * self.delta_t + self.sigma_sp * self.Brownian_motion()
                )
                y_new = (
                    y + delta_y * self.delta_t + self.sigma_sp * self.Brownian_motion()
                )
                op_new = (
                    op
                    + self.opinion_interaction(i) * self.delta_t
                    + self.sigma_op * self.Brownian_motion()
                )

                # add terms for potentials
                # potential landscape
                if self.useDoublewell:
                    x_new -= self.partialX(x) * self.delta_t
                    y_new -= self.partialY(y) * self.delta_t
                new_x[i] = x_new
                new_y[i] = y_new
                new_op[i] = op_new
            for i in range(self.N):
                # update agents
                self.agents[i].set_x(new_x[i])
                self.agents[i].set_y(new_y[i])
                self.agents[i].set_opinion(new_op[i])

            self.steps += 1

            # update distances and clustering
            self.calculate_distances()
            if self.clustering:
                self.cluster()
            if self.datasim and self.bound:
                self.boundary()
            if self.datasim:
                self.gridcell = [
                    self.get_grid_cell(coord) for coord in self.get_coordinates()
                ]

            # update snapshot information list
            coords[step] = pd.DataFrame(self.get_coordinates())
            opinions[step] = self.get_opinions()

            # networks[step] = self.get_network()
            networks[step] = self.get_adjacency_matrix()
            labels[step] = self.labels
            gridcell[step] = self.gridcell
        # append simulation snapshots to results
        self.results["Coordinates"] = self.results["Coordinates"] + coords
        self.results["Opinions"] = pd.concat(
            [self.results["Opinions"], pd.DataFrame(opinions).transpose()],
            axis=1,
            ignore_index=True,
        )

        self.results["Networks"] = self.results["Networks"] + networks
        self.results["Cluster Labels"] = pd.concat(
            [self.results["Cluster Labels"], pd.DataFrame(labels).transpose()],
            axis=1,
            ignore_index=True,
        )
        self.results["gridcell"] = pd.concat(
            [self.results["gridcell"], pd.DataFrame(gridcell).transpose()],
            axis=1,
            ignore_index=True,
        )

    def cluster(self, eps=None, minpoints=None):
        """
        clusters the agents using DBSCAN or HDBSCAN, if clustering is enabled

        Parameters
        ----------
        eps : float, optional
            eps parameter used for DBSCAN/HDBSCAN clustering. The default is None, resulting in the usage of the class variable eps.
        minpoints : int, optional
            min_samples parameter for DBSCAN clustering. The default is None, resulting in the usage of the class variable minpoints.

        Returns
        -------
        None.

        """

        # Allow clustering based on spatial data only, opinion data only or a combination of both. This is defined by the class variables
        # define the correct datastructre for clustering
        if self.spatial_clustering and self.opinion_clustering:
            opinions = np.array(self.get_opinions())
            coords = np.array(self.get_coordinates())
            # opinions has to be reshaped before appending:
            # opinions[None] transforms the shape from (300,) to (1,300) and .T to (300,1)
            coords = np.append(coords, opinions[None].T, axis=1)
        elif self.spatial_clustering:
            coords = np.array(self.get_coordinates())
        elif self.opinion_clustering:
            coords = np.array(self.get_opinions())
            coords = coords.reshape(-1, 1)

        # calculate clusters
        if not self.useHDBSCAN:
            ### DBSCAN
            if eps == None:
                eps = self.eps
            if minpoints == None:
                minpoints = self.minpoints
            if self.dynamic_eps:
                self.eps, bounded = self.calculate_eps()
                if bounded:  # calculated eps is larger than the interaction radius
                    # eps is set to a value slightly larger than the smallest interaction radius, allowing agents which are slightly outside of interaction range to cluster
                    self.eps += 0.1 * min(self.R_op, self.R_sp)

            db = DBSCAN(eps=self.eps, min_samples=minpoints).fit(coords)
            self.labels = db.labels_
        else:
            ### HDBSCAN
            hdb = hdbscan.HDBSCAN(min_cluster_size=self.min_cluster_size).fit(coords)
            self.labels = hdb.labels_

        # update number of clusters and see if there are unclustered agents
        self.hasNoise = 1 if -1 in self.labels else 0
        self.n_clusters = len(set(self.labels)) - self.hasNoise

    def calculate_eps(self, minpoints=None):
        """
        calculates the knee point of the k-neighbours graph, where k is the minimum number of points/agents in a cluster. This value is considered to be the optimal value for eps. The calculated eps shall not be larger than the opinion radius or the spatial radius.

        Parameters
        ----------
        minpoints : int, optional
            min_samples parameter for DBSCAN clustering. The default is None, resulting in the usage of the class variable minpoints.. The default is None.

        Returns
        -------
        min_eps : float
            The minimum of the calculated eps, opinion radius and spatial radius.
        bounded : bool
            specifies if either the opinion radius or the spatial radius was smaller than the calculated eps.

        """

        if minpoints == None:
            minpoints = self.minpoints
        else:
            minpoints = minpoints

        # define the correct datastructre for nearest neighbor calculation
        if self.spatial_clustering and self.opinion_clustering:
            opinions = np.array(self.get_opinions())
            coords = np.array(self.get_coordinates())
            np.append(
                coords, opinions[None].T, axis=1
            )  # opinions[None] transforms the shape from (300,) to (1,300) and .T to (300,1)
        elif self.spatial_clustering:
            coords = np.array(self.get_coordinates())
        elif self.opinion_clustering:
            coords = np.array(self.get_opinions())

        # calculate k nearest neighbors distances
        nearest_neighbors = NearestNeighbors(n_neighbors=minpoints + 1)
        neighbors = nearest_neighbors.fit(coords)
        distances, indeces = neighbors.kneighbors(coords)

        # find optimal eps
        distances = np.sort(distances[:, minpoints], axis=0)
        i = np.arange(len(distances))
        knee = KneeLocator(
            i,
            distances,
            S=1,
            curve="convex",
            direction="increasing",
            interp_method="polynomial",
        )

        # update eps
        eps = distances[knee.knee] + 0.2 * distances[knee.knee]
        min_eps = min(eps, self.R_op, self.R_sp)
        bounded = False if min_eps == eps else True

        return (min_eps, bounded)

    def calculate_distances(self):
        """
        calculates all pairwise distances between current agents and updates the distance matrix

        Returns
        -------
        None.

        """

        coordinates = self.get_coordinates()
        self.distancematrix = cdist(coordinates, coordinates, "euclidean")

    def Brownian_motion(self):
        """
        generate brownian motion

        Returns
        -------
        A value emulating brownian motion
        """
        return np.sqrt(self.delta_t) * self.rng.normal(loc=0, scale=1)

    def spatial_interaction(self, k):
        """
        defines interaction map influencing spatial movement(U tilde)

        Parameters
        ----------
        k : int
            index of agent.

        Returns
        -------
        float
            change to x coordinate of agent k from spatial interactions.
        float
            change to y coordinate of agent k from spatial interactions.

        """

        x = 0
        y = 0

        for j in range(self.N):
            x_j, y_j = self.spatial_pairwise_interaction(k, j)
            x += x_j
            y += y_j
        if self.normalize_with_neighbour_num:
            neighbours = sum(self.get_adjacency_matrix()[k, :])
            if neighbours == 0:
                neighbours = 1
            return x / neighbours, y / neighbours
        return x / self.N, y / self.N

    def opinion_interaction(self, k):
        """
        defines interaction map influencing opinion states (V tilde)

        Parameters
        ----------
        k : int
            index of agent.

        Returns
        -------
        float
            change to opinion of agent k from spatial interactions.

        """

        res = 0

        for j in range(self.N):
            res += self.opinion_pairwise_interaction(k, j)

        if self.normalize_with_neighbour_num:
            neighbours = sum(self.get_adjacency_matrix(useR_sp=False)[k, :])
            if neighbours == 0:
                neighbours = 1
            return res / neighbours
        return res / self.N

    def spatial_pairwise_interaction(self, k, j):
        """
        defines pairwise interaction map (U)

        Parameters
        ----------
        k : int
            index of first agent.
        j : int
            index of second.

        Returns
        -------
        float
            change to x coordinate of agent k from pairwise spatial interaction to agent j.
        float
            change to y coordinate of agent k from pairwise spatial interaction to agent j..

        """

        if self.distancematrix[k, j] <= self.R_sp:
            temp = self.beta * np.sign(
                self.agents[j].get_opinion() * self.agents[k].get_opinion()
            )
            return (
                temp * (self.agents[j].get_x() - self.agents[k].get_x()),
                temp * (self.agents[j].get_y() - self.agents[k].get_y()),
            )
        else:
            return (0, 0)

    def opinion_pairwise_interaction(self, k, j):
        """
        defines pairwise interaction map (V)

        Parameters
        ----------
        k : int
            index of first agent.
        j : int
            index of second.

        Returns
        -------
        float
            change to opinion of agent k from pairwise spatial interaction to agent j.
        """
        if self.distancematrix[k, j] <= self.R_op:
            return self.alpha * (
                self.agents[j].get_opinion() - self.agents[k].get_opinion()
            )
        else:
            return 0

    def get_results(self):
        """
        returns the dictionary containing the simulation

        Returns
        -------
        dictionary
            Dictionary containing "Coordinates", "Opinions",
            "Networks", "Cluster Labels" and "gridcell".

            "Coordinates": list of pandas DataFrames
                Each DataFrame contains the x coordinates of agents in the first column and y
                coordinates of agents in the second column. The list index corresponds to the
                simulation snapshot.

            "Opinions": pandas DataFrame
                Rows represent agents and columns represent simulation snapshots. The entries are
                the opinion values.


            "Networks": list of numpy matrices
                The list index corresponds to the simulation snapshot. Each entry is the adjacency
                matrix, describing the network, for the snapshot.

            "Cluster Labels": pandas DataFrame
                Rows represent agents and columns represent simulation snapshots.
                The entries are the cluster label, with -1 denoting noise.

            "gridcell": pandas DataFrame
                Rows represent agents and columns represent simulation snapshots.
                The entries are the gridcell the agent is located in.
                The cells of the 3x3 grid are enumerated as follows:
                    6 7 8
                    3 4 5
                    0 1 2
        """
        return self.results

    def get_opinions(self):
        """
        returns the opinion of all agents as a list

        Returns
        -------
        list
            opinions of all agents.

        """

        return [agent.get_opinion() for agent in self.agents]

    def get_coordinates(self):
        """
        returns the coordinates of all agents as a list of lists

        Returns
        -------
        list of lists
            coordinates of all agents [[x1,y1],[x2,y2]...].
        """

        return [agent.get_coordinates() for agent in self.agents]

    def get_adjacency_matrix(
        self, useR_sp: bool = True, time_step: Optional[int] = None
    ):
        """
        Returns the adjacency matrix of the latest snapshot. Two agents are adjacent to each other
        if the distance between them is smaller than the spatial radius.

        Parameters
        ----------
        @param useR_sp : bool, optional
            specifies if the spatial radius should be used to determine adjacency.
            The default is True.
        @param useR_sp : Bool,optional
            If false, the opinion Radius will be used for the network. The default is True
        @param time_step : int, optional
            for which timestep the adjacency matrix is computed.
            The default is the end of the simulation.

        Returns
        -------
        res : numpy matrix
            adjacency matrix, describing the edges of the network. The rows and columns of the
            matrix represent the agents. Entries of the matrix are 0 and 1 indicating no edge and
            edge, respectively.

        """
        radius = self.R_sp if useR_sp else self.R_op
        if time_step is None:
            res = np.zeros((self.N, self.N), dtype=int)
            res[self.distancematrix <= radius] = 1
            # set the diagonal to zero
            np.fill_diagonal(res, 0)
            return res
        else:
            coordinates = self.get_results()["Coordinates"][time_step]
            distances = cdist(coordinates, coordinates, "euclidean")
            res = np.zeros((self.N, self.N), dtype=int)
            res[distances <= radius] = 1
            # set the diagonal to zero
            np.fill_diagonal(res, 0)
            return res

    def get_degree_distributions(
        self,
        time_step: Optional[int] = None,
        bins: Optional[Union[np.ndarray, int]] = None,
    ) -> dict[str, tuple[int, int]]:
        """
        Returns the degree distribution of the network defined by the spatial and opinion
        interaction function respectively. The network corresponds to the final timestep of the model.


        Returns
        -------
        dict :
        containing the spatial degree distribution with key 'spatial' as numpy ndarray
        and the opinion degree distribution with key 'opinion' as numpy ndarray. The first entry of each array contains
        the counts and the second entry the bin edges.
        """
        if bins is None:
            # The default is one bin for every possible degree
            bins = np.arange(-0.5, self.N + 1.5, 1)

        spatial_degrees = np.sum(
            self.get_adjacency_matrix(useR_sp=True, time_step=time_step), axis=1
        )
        opinion_degrees = np.sum(
            self.get_adjacency_matrix(useR_sp=False, time_step=time_step), axis=1
        )
        return {
            "Spatial": np.histogram(spatial_degrees, bins),
            "Opinion": np.histogram(opinion_degrees, bins),
        }

    def get_clusters(self):
        """
        returns the list of labels which show the cluster-membership of all agents

        Returns
        -------
        list
            list of cluster labels.

        """

        return self.labels

    def get_data(self):
        """
        returns a dataframe containing coordinates, opinions and gridcell

        Returns
        -------
        list
        """
        coords = self.get_coordinates()
        opinions = self.get_opinions()

        data = pd.DataFrame(coords, columns=["x", "y"])
        data["opinions"] = opinions
        data["gridcell"] = self.gridcell
        return data

    def get_results_as_numpy_array(self):
        """Converts the results to a numpy array So that it can be saved more efficently.
        We are ommiting some data, like adjacency matrix, clustering, grid cells etc."""
        data = [
            [
                *self.results["Coordinates"][step].to_numpy().transpose(),
                self.results["Opinions"].iloc[:, step],
            ]
            for step in range(self.steps + 1)
        ]
        return np.array(data)

    def set_results_from_numpy_array(self, array, cluster=False):
        """Brings loaded data into the appropriate data formats."""
        assert isinstance(array, np.ndarray)

        # Extract coordinates
        coords = [pd.DataFrame(array[i][:2].T) for i in range(array.shape[0])]

        # Extract opinions and convert to DataFrame
        opinions = pd.DataFrame(array[:, 2]).T

        # Assign the results to the self.results dictionary
        self.results["Opinions"] = opinions
        self.results["Coordinates"] = coords

        # Storing the adjacency matrix of the final time step of the simulation.
        final_coordinates = array[-1]
        self.distancematrix = cdist(
            np.array([final_coordinates[0], final_coordinates[1]]).T,
            np.array([final_coordinates[0], final_coordinates[1]]).T,
            "euclidean",
        )
        x, y = final_coordinates[:2]
        for i, agent in enumerate(self.agents):
            agent.set_opinion(opinions[self.steps][i])
            agent.set_y(y[i])
            agent.set_x(x[i])

        if cluster:
            # This will only set self.labels and self.n_clusters and not fill the entire
            # resutls dictionary
            self.cluster()

    def get_grid_cell(self, coordinate):
        """
        calculates the gridcell, given a coordinate.
        The cells of the 3x3 grid are enumerated as follows:
            6 7 8
            3 4 5
            0 1 2

        Parameters
        ----------
        coordinate : tuple of floats
            x and y coordinate (x,y).

        Returns
        -------
        int
            gridcell in which the coordinate is located at.
        None
            if the coordinate is outside of the grid
        """

        x, y = coordinate[0], coordinate[1]

        if x < self.xlim[0] or x > self.xlim[1] or y < self.ylim[0] or y > self.ylim[1]:
            print("Coordinate outside the grid")
            return None  # Coordinate outside the grid

        cell_width = (
            self.xlim[1] - self.xlim[0]
        ) / self.ngrid  # Assuming 3 cells horizontally
        cell_height = (
            self.ylim[1] - self.ylim[0]
        ) / self.ngrid  # Assuming 3 cells vertically

        if x == self.xlim[0]:
            cell_x = 0
        elif x == self.xlim[1]:
            cell_x = 2
        else:
            cell_x = int(
                (x - self.xlim[0]) // cell_width
            )  # calculate column of the grid
        if y == self.ylim[0]:
            cell_y = 0
        elif y == self.ylim[1]:
            cell_y = 2
        else:
            cell_y = int((y - self.ylim[0]) // cell_height)  # calculate row of the grid

        return cell_x + cell_y * 3  # transform column and row information to grid cell

    def set_useDoubleWell(self, useDoublewell):
        """
        sets the value of the class variable useDoublewell

        Parameters
        ----------
        useDoublewell : bool
            Value the class variable useDoublewell is set to.

        Returns
        -------
        None.

        """
        self.useDoublewell = useDoublewell

    def global_assortativity(
        self, time_step: Optional[int] = None, use_spatial_radius: bool = True
    ) -> float:
        if time_step is None:
            time_step = self.steps
        # initialize thee data
        A = self.get_adjacency_matrix(useR_sp=use_spatial_radius, time_step=time_step)
        opinions = self.get_results()["Opinions"][time_step].to_numpy()
        # calculate degree sequence and number of edges
        degrees = np.sum(A, axis=0)
        num_edges = np.sum(A) / 2
        # degree weighted mean
        deg_mean = np.dot(degrees, opinions) / (2 * num_edges)
        # deg weighted standard deviation
        deg_sigma = np.sqrt(
            np.dot(degrees, (opinions - deg_mean) ** 2) / (2 * num_edges)
        )
        attributes = (opinions - deg_mean) / deg_sigma
        return (attributes @ A @ attributes) / (2 * num_edges)

    def compute_borders(self):
        """
        computes the borders of the gridcells
        class variable borders is a dictionary containing a 4-tuple for each gridcell.
        the 4-tuple describes the cell boundaries (left, right, lower, upper)

        Returns
        -------
        None.

        """
        # generate 4 equidistant values, resulting in a split into 3 sections
        border_x = np.round(np.linspace(self.xlim[0], self.xlim[1], self.ngrid + 1), 5)
        border_y = np.round(np.linspace(self.ylim[0], self.ylim[1], self.ngrid + 1), 5)

        for n1 in range(3):
            for n2 in range(3):
                self.borders[n1 * 3 + n2] = (
                    border_x[n2],
                    border_x[n2 + 1],
                    border_y[n1],
                    border_y[n1 + 1],
                )

    def boundary(self):
        """
        Enforces the boundary conditions (only for the outer boundaries of the domain)

        Returns
        -------
        None.

        """
        coords = self.get_coordinates()
        # iterate over all agents
        for i, coord, cell in zip(range(self.N), coords, self.gridcell):
            if cell == 4:
                continue
            # get the correct cell borders
            left, right, bottom, top = self.borders[cell]
            resets = 0
            # if agent is outside of the borders the agent is moved back to the border
            if coord[0] < left and cell in (0, 3, 6):
                self.agents[i].set_x(left)
                resets += 1
            if coord[0] > right and cell in (2, 5, 8):
                self.agents[i].set_x(right)
                resets += 1
            if coord[1] < bottom and cell in (0, 1, 2):
                self.agents[i].set_y(bottom)
                resets += 1
            if coord[1] > top and cell in (6, 7, 8):
                self.agents[i].set_y(top)
                resets += 1
        # print(resets/self.N)

    def plot(self, *args, **kwargs) -> plt.Axes:
        plotter = ModelPlotter(self)
        return plotter.plot(*args, **kwargs)

    def plot_with_marginals(
        self,
        *args,
        **kwargs,
    ) -> Optional[plt.Figure]:

        return ModelPlotter(self).plot_with_marginals(*args, **kwargs)

    def plot_opinions(
        self,
        *args,
        **kwargs,
    ) -> plt.Axes:

        return ModelPlotter(self).plot_opinions(*args, **kwargs)

    def double_well_potential(self, x: float, y: float) -> float:
        """
        Definition of the double well potential for plotting.

        Parameters
        ----------
        x : int
            x-coordinate.
        y : int
            y-coordinate.

        Returns
        -------
        int
            double-well potential at the coordinate.

        """
        return 3 * (x**2 - 0.25) ** 2 + y**2

    def partialX(self, x: float) -> float:
        return 12 * x * (x**2 - 0.25)

    def partialY(self, y: float) -> float:
        return 2 * y


# I used this area to play around and test. Nothing important is happening below here.

if __name__ == "__main__":
    pass
