import warnings
from typing import Optional, Iterable, Union

import matplotlib
import matplotlib.lines as mlines
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable

from src import plotting_aux_file as aux

plt.rcParams["text.usetex"] = True


class ModelPlotter:
    def __init__(self, model):
        self.model = model

    def plot(
        self,
        timestep: Optional[int] = None,
        xlim: Optional[tuple[float, float]] = None,
        ylim: Optional[tuple[float, float]] = None,
        print_title: Optional[bool] = True,
        show_network: bool = False,
        use_spatial_network: bool = True,
        show_clusters: bool = False,
        ax: Optional[plt.axes] = None,
        show_landscape: bool = False,
        landscape_colorbar=True,
        cmap: str | matplotlib.colors.Colormap = "turbo",
        op_cmap: str | matplotlib.colors.Colormap = aux.discontinious_red_blue,
        node_color: Optional[Iterable] = None,
        colorbar_left: bool = False,
        colorbar: bool = True,
        colorbar_ticks: Optional[Iterable] = None,
        colorbar_fontsize: Optional[float] = None,
        cluster_legend_outside: bool = True,
        opinion_label: str = "Opinion",
        s: Optional[float] = None,
        network_alpha: Optional[float] = None,
        network_lw: Optional[float] = None,
        node_linewidth: float = None,
    ) -> plt.Axes:
        """
        Plots the agents in the social space.

        Parameters
        ----------
        @param timestep : int, optional
            timestep/snapshot of the simulation which is to be plotted. The default is None, resulting in plotting of
            the latest snapshot of the simulation.
        @param xlim : tuple of floats, optional
            interval of the x-axis (lower bound, upper bound) of the figure. The default is None.
        @param ylim : tuple of floats, optional
            interval of the y-axis (lower bound, upper bound) of the figure. The default is None.
        @param print_title : bool, optional
            specifies if the title should be printed. The default is True.
        @param show_network : bool, optional
            specifies if the network edges should be shown. The default is False.
        @param use_spatial_network : bool, optional
            The default is True. If False, the network induced by R_op is plotted
        @param network_alpha: float, optional
            Sets the opacity for the network. The Default is None.
        @param network_lw: str, optional
            Sets the line width for the edges in the network.
            The default is None and will use the default from networkx.
        @param show_clusters : bool, optional
            specifies if cluster membership should be shown. The default is False.
        @param ax : Axes object from matplotlib, optional
            object on which the graph is plotted. The default is None, resulting in the creation of a new Axes object.
        @param show_landscape : bool, optional
            specifies if the double well potential should be shown. The default is False.
        @param cmap : str, optional
            specifies which color map should be used for the potential landscape (parula, turbo).
             The default is "turbo".
        @param op_cmap : str, optional
            specifies which color map should be used for the opinions. The default is the red-blue colormap with
             a discontinuity at zero. The custom colormap can be found in the plotting_functions.py file.
        @param s: float, optional
            The size of the markers used for the agent in the social space. The dfault is None and will use the
            default size in matplotlib.pyplot.rcParams
        @param colorbar_left: bool, optional
            Will move the colorbar to the left of the figure. The default is False. This is important, when plotting the
            marginal distribution of agents in the social space.
        @param opinion_label: str, optional
            Sets the label for the colorbar. The default is "Opinion"
        @param cluster_legend_outside: bool, optional
            If True the legend for the cluster marker will be located outside the figure. Otherwise, the legend
            will be located in the figure and may overlap with the plot.
        @param colorbar_fontsize: float, optional
            Sets the font size for the colorbar labels.
        @param colorbar_ticks: Iterable
            Set the ticks for the colorbar. Must contain floats.
        @param colorbar: bool
            Determines if the colorbar will be plotted.

        Raises
        ------
        ValueError
            Raised if cmap is not "parula" or "turbo".

        Returns
        -------
        An Axes object from matplotlib containing the plot.

        """

        if timestep is None:
            timestep = self.model.steps

        coords = self.model.results["Coordinates"][timestep]
        opinions = self.model.results["Opinions"][timestep]
        cluster_labels = self.model.results["Cluster Labels"][timestep]

        if node_color is None:
            node_color = opinions
            colorbar_limits = (self.model.opdim[0], self.model.opdim[1])
        else:
            colorbar_limits = (min(node_color), max(node_color))

        if ax is None:
            ax = plt.gca()

        # predefine the locations of the colorbar
        divider = make_axes_locatable(plt.gca())
        if show_clusters and cluster_legend_outside:
            lax = divider.append_axes("bottom", "5%", pad="3%")
        if show_landscape and landscape_colorbar:
            cpot_ax = divider.append_axes("left", "5%")

        # draw network
        if show_network:
            if network_lw is None:
                network_lw = 1
            if network_alpha is None:
                network_alpha = 0.1

            if use_spatial_network:
                # define the network with the spatial radius
                sp_adjacency_matrix = self.model.get_adjacency_matrix(
                    useR_sp=True, time_step=timestep
                )
                network = nx.from_numpy_array(sp_adjacency_matrix)
            else:
                # define the network with the opinion radius
                op_adjacency_matrix = self.model.get_adjacency_matrix(
                    useR_sp=False, time_step=timestep
                )
                network = nx.from_numpy_array(op_adjacency_matrix)
            # draw the network
            line_style = "solid" if use_spatial_network else "-."
            nx.draw_networkx_edges(
                network,
                coords.to_numpy(),
                alpha=network_alpha,
                ax=ax,
                width=network_lw,
                style=line_style,
            )
            # Proper label settings
            ax.tick_params(left=True, bottom=True, labelleft=True, labelbottom=True)

        # draw clustered agents with different markers
        if show_clusters:
            if None in cluster_labels.values:
                warnings.warn(
                    "Agents have not been clustered. Set the clustering algorithm in Model.__init__"
                )
                sc = ax.scatter(
                    x=coords[0],
                    y=coords[1],
                    c=node_color,
                    cmap=op_cmap,
                    edgecolors="black",
                    vmin=colorbar_limits[0],
                    vmax=colorbar_limits[1],
                )
            else:
                handles = []
                hasNoise = -1 in cluster_labels.values
                n_clusters = len(set(cluster_labels)) - hasNoise
                try:
                    aux.marker_list[n_clusters]
                except IndexError:
                    raise ValueError(
                        f"To many clusters were found to plot them. The number of clusters is above{len(aux.marker_list)}"
                    )
                if not all(node_color == opinions):
                    raise NotImplementedError(
                        "With clustering the nodes have to be colored by"
                        "their opinions."
                    )

                for label in range(
                    n_clusters
                ):  # iterate over clusters and plot with corresponding shape
                    temp = coords[
                        cluster_labels.values == label
                    ]  # slice only the datapoints that belong to current cluster
                    temp_opinion = opinions[cluster_labels.values == label]
                    sc = ax.scatter(
                        temp.iloc[:, 0],
                        temp.iloc[:, 1],
                        label=label - 1,
                        c=temp_opinion,
                        cmap=op_cmap,
                        vmin=self.model.opdim[0],
                        vmax=self.model.opdim[1],
                        marker=aux.marker_list[label],
                        edgecolor="black",
                        linewidth=0.5,
                    )
                    handles.append(
                        mlines.Line2D(
                            [],
                            [],
                            color="black",
                            marker=aux.marker_list[label],
                            ls="",
                            label=label,
                        )
                    )
                if hasNoise:  # plotting of noise agents
                    if not all(node_color == opinions):
                        raise NotImplementedError(
                            "With clustering the nodes have to be colored by"
                            "their opinions."
                        )
                    temp = coords[
                        cluster_labels.values == -1
                    ]  # slice only the datapoints that belong to current cluster
                    temp_opinion = opinions[cluster_labels.values == -1]
                    sc = ax.scatter(
                        temp.iloc[:, 0],
                        temp.iloc[:, 1],
                        label="Noise",
                        c=temp_opinion,
                        cmap=op_cmap,
                        vmin=self.model.opdim[0],
                        vmax=self.model.opdim[1],
                        marker=".",
                        edgecolor="black",
                        linewidth=0.5,
                    )
                    handles.append(
                        mlines.Line2D(
                            [], [], color="black", marker=".", ls="", label="Noise"
                        )
                    )
                if cluster_legend_outside:
                    lax.axis("off")
                    lax.legend(
                        handles=handles,
                        bbox_to_anchor=(0.5, -0.05),
                        loc="upper center",
                        title="Clusters",
                        ncol=min(n_clusters + hasNoise, 5),
                    )
                else:
                    ax.legend(
                        handles=handles,
                        loc="upper left",
                        title="Clusters",
                        ncols=1,
                        borderpad=0.1,
                    )
        else:
            if s is None:
                # set the marker size for the agents
                s = 0.5 * matplotlib.rcParams["lines.markersize"] ** 2
            # Plot the agents
            sc = ax.scatter(
                x=coords[0],
                y=coords[1],
                c=node_color,
                cmap=op_cmap,
                edgecolors="black",
                vmin=colorbar_limits[0],
                vmax=colorbar_limits[1],
                s=s,
                linewidths=node_linewidth,
            )
        # placing the opinion colorbar
        if colorbar_left and colorbar:
            # creates a new axis for the colorbar
            cax = divider.append_axes("left", "5%", pad="28%")
            plt.colorbar(
                sc,
                orientation="vertical",
                label=opinion_label,
                cax=cax,
                ax=sc,
                ticklocation="left",
            )
        elif colorbar:
            # creates a new axis for the colorbar
            cax = divider.append_axes("right", "5%", pad="3%")
            cbar = plt.colorbar(sc, orientation="vertical", cax=cax, ax=sc)
            if colorbar_ticks is not None and colorbar_fontsize is not None:
                cbar.set_ticks(
                    colorbar_ticks,
                    labels=[r"$" + str(tick) + r"$" for tick in colorbar_ticks],
                    fontsize=colorbar_fontsize,
                )
                cbar.set_label(label=opinion_label, fontsize=colorbar_fontsize)

        if print_title:
            ax.set_title(
                r"Model Snapshot at time $t ={:.2f}$".format(
                    timestep * self.model.delta_t
                )
                + "\n"
                + r"for $N ={}, R_{{op}} ={}, R_{{sp}} ={}, \alpha ={}, \beta ={},"
                r"\sigma_{{op}} ={}, \sigma_{{sp}} ={}$".format(
                    self.model.N,
                    self.model.R_op,
                    self.model.R_sp,
                    self.model.alpha,
                    self.model.beta,
                    self.model.sigma_op,
                    self.model.sigma_sp,
                )
            )
        if xlim is not None:
            if max(coords[0]) > xlim[1] or min(coords[0]) < xlim[0]:
                warnings.warn("There are datapoints outside the range of xlim")
            ax.set_xlim(xlim)
        if ylim is not None:
            if max(coords[1]) > ylim[1] or min(coords[1]) < ylim[0]:
                warnings.warn("There are datapoints outside the range of xlim")
            ax.set_ylim(ylim)

        # draw potential landscape
        if show_landscape:
            cmaps = {"parula", "turbo"}
            # calculate potential landscape
            xmin, xmax, ymin, ymax = ax.axis()
            dx, dy = 0.01, 0.01
            x = np.arange(xmin, xmax, dx)
            y = np.arange(ymin, ymax, dy)
            X, Y = np.meshgrid(x, y)
            Z2 = self.model.double_well_potential(X, Y)
            if cmap == "parula":
                cmap_potential = aux.parula
            elif cmap == "turbo":
                cmap_potential = "turbo"
            else:
                warnings.warn(
                    f"{cmap} is not a valid colormap for the potential landscape"
                )
            # draw landscape and colorbar
            pot = ax.imshow(
                Z2,
                cmap=cmap_potential,
                alpha=0.9,
                interpolation="bilinear",
                extent=ax.axis(),
            )
            if landscape_colorbar:
                plt.colorbar(
                    pot,
                    orientation="vertical",
                    label="Potential",
                    cax=cpot_ax,
                    ticklocation="left",
                )
            # plt.colorbar(pot, orientation = "vertical", label = "Potential", location = "left")
        return ax

    def plot_with_marginals(
        self,
        timestep: Optional[int] = None,
        bins: int | np.ndarray = 31,
        show_landscape: bool = False,
        show_clusters: bool = False,
        show_network: bool = True,
        use_spatial_network: bool = True,
        print_title: bool = True,
        xlim: Optional[tuple[float, float]] = None,
        ylim: Optional[tuple[float, float]] = None,
        filename: Optional[str] = None,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
        opinion_label: Optional[str] = None,
        fig=Optional[plt.Figure],
        no_ticks: bool = False,
        colorbar: bool = True,
        s: Optional[float] = None,
        network_alpha: Optional[float] = None,
        network_lw: Optional[float] = None,
        hist_ylim: int = None,
        node_linewidth=None,
    ) -> plt.Figure:
        """
        Create a plot of the agents in the social space with the marginal distributions plotted next to the social
        space. The colors of the histogram bins correspond to the average opinion of the agents that contribute to
        that bin.
        @param timestep : int, optional
            timestep/snapshot of the simulation which is to be plotted. The default is None, resulting in plotting of
            the latest snapshot of the simulation.
        @param bins: int, numpy.ndarray, optional
            Set the bin_edegs directly or the number of bin edges. So for 5 bins set 6.
        @param hist_ylim: int, optional
            set the upper limit for the histograms. The lower Limit is 0. The default is the maximum count of both
            hitograms
        @param filename: str, optional
            If specified, the plot will be saved here. Include the file ending and path to directoty if needed.
        @param xlabel:str, optional
            Label for x-axis
        @param ylabel:str, optional
            Label for y-axis
        @param fig: Figure object from Matplotlib, optional,
            object on which the graph is plotted. The default is None, resulting in the creation of a new Figure object.
        @param no_ticks: bool
            remove the numbering on the x-axis and y-axis of the social space
        @param xlim : tuple of floats, optional
            interval of the x-axis (lower bound, upper bound) of the figure. The default is None.
        @param ylim : tuple of floats, optional
            interval of the y-axis (lower bound, upper bound) of the figure. The default is None.
        @param print_title : bool, optional
            specifies if the title should be printed. The default is True.
        @param show_network : bool, optional
            specifies if the network edges should be shown. The default is False.
        @param use_spatial_network : bool, optional
            The default is True. If False, the network induced by R_op is plotted
        @param network_alpha: float, optional
            Sets the opacity for the network. The Default is None.
        @param network_lw: str, optional
            Sets the line width for the edges in the network.
            The default is None and will use the default from networkx.
        @param show_clusters : bool, optional
            specifies if cluster membership should be shown. The default is False.
        @param show_landscape : bool, optional
            specifies if the double well potential should be shown. The default is False.
        @param cmap : str, optional
            specifies which color map should be used for the potential landscape (parula, turbo).
             The default is "turbo".
        @param op_cmap : str, optional
            specifies which color map should be used for the opinions. The default is the red-blue colormap with
             a discontinuity at zero. The custom colormap can be found in the plotting_functions.py file.
        @param s: float, optional
            The size of the markers used for the agent in the social space. The dfault is None and will use the
            default size in matplotlib.pyplot.rcParams
        @param colorbar_left: bool, optional
            Will move the colorbar to the left of the figure. The default is False. This is important, when plotting the
            marginal distribution of agents in the social space.
        @param opinion_label: str, optional
            Sets the label for the colorbar. The default is "Opinion"
        @param cluster_legend_outside: bool, optional
            If True the legend for the cluster marker will be located outside the figure. Otherwise, the legend
            will be located in the figure and may overlap with the plot.
        @param colorbar_fontsize: float, optional
            Sets the font size for the colorbar labels.
        @param colorbar_ticks: Iterable
            Set the ticks for the colorbar. Must contain floats.
        @param colorbar: bool
            Determines if the colorbar will be plotted.

        @return: plt.Figure object
        """
        if timestep is None:
            timestep = self.model.steps
        # fetching the data
        x, y = self.model.get_results()["Coordinates"][timestep].to_numpy().transpose()
        opinions = self.model.get_results()["Opinions"][timestep]

        if fig is None:
            fig = plt.figure()
        # Create the main axes, leaving 25% of the figure space at the top and on the
        # right to position marginals.
        ax = fig.add_gridspec(top=0.75, right=0.75).subplots()
        # The main axes' aspect can be fixed.
        ax.set(aspect=1)
        # Create marginal axes, which have 25% of the size of the main axes.  Note that
        # the inset axes are positioned *outside* (on the right and the top) of the
        # main axes.
        ax_histx = ax.inset_axes((0, 1.05, 1, 0.25), sharex=ax)
        ax_histy = ax.inset_axes((1.05, 0, 0.25, 1), sharey=ax)

        # setting the labels at the correct position
        ax_histx.tick_params(axis="x", labelbottom=False)
        ax_histy.tick_params(axis="y", labelleft=False)

        ax_histy.grid(True)
        ax_histx.grid(True)

        # the scatter plot:
        self.plot(
            timestep=timestep,
            xlim=xlim,
            ylim=ylim,
            print_title=print_title,
            show_network=show_network,
            use_spatial_network=use_spatial_network,
            show_clusters=show_clusters,
            ax=ax,
            show_landscape=show_landscape,
            colorbar_left=True,
            opinion_label=opinion_label,
            colorbar=colorbar,
            s=s,
            network_alpha=network_alpha,
            network_lw=network_lw,
            node_linewidth=node_linewidth,
        )

        if no_ticks:
            ax.xaxis.set_ticklabels([])
            ax.yaxis.set_ticklabels([])

        # now determine nice limits by hand:
        if xlim is None:
            xlim = ax.get_xlim()
        if ylim is None:
            ylim = ax.get_ylim()
        if isinstance(bins, int):
            bins = np.linspace(*xlim, bins)

        # determine the coloring for the histograms
        av_opinion_x = np.zeros(len(bins) - 1)
        av_opinion_y = np.zeros(len(bins) - 1)
        for k, _ in enumerate(zip(av_opinion_x, av_opinion_y)):
            opx = np.mean(opinions[(bins[k] < x) & (bins[k + 1] > x)])
            opy = np.mean(opinions[(bins[k] < y) & (bins[k + 1] > y)])
            if not np.isnan(opx):
                av_opinion_x[k] = opx
            if not np.isnan(opy):
                av_opinion_y[k] = opy

        x_counts, _ = np.histogram(x, bins=bins)
        y_counts, _ = np.histogram(y, bins=bins)

        # Changing the data type, otherwise ax.bar raises an error
        x_counts = np.array(x_counts, dtype=float)
        y_counts = np.array(y_counts, dtype=float)

        # We neet to supply color values from the interval [0, 1] to the colormap, so this is just linear rescaling.
        rescale = lambda y: (y + 1) / 2
        ax_histx.bar(
            x=bins[:-1] + np.diff(bins) / 2,
            height=x_counts,
            width=bins[1] - bins[0],
            color=aux.discontinious_red_blue(rescale(av_opinion_x)),
            edgecolor="k",
        )
        ax_histy.barh(
            y=bins[:-1] + np.diff(bins) / 2,
            width=y_counts,
            height=bins[1] - bins[0],
            color=aux.discontinious_red_blue(rescale(av_opinion_y)),
            edgecolor="k",
        )

        if hist_ylim is None:
            # Set the maximum count as limit for the histogram plots
            ax_histx.set_ylim(0, max(np.concatenate([x_counts, y_counts])) + 1)
            ax_histy.set_xlim(0, max(np.concatenate([x_counts, y_counts])) + 1)
        else:
            ax_histx.set_ylim(0, hist_ylim)
            ax_histy.set_xlim(0, hist_ylim)

        # Setting the Zorder of gridlines and ticks.
        ax_histx.set_axisbelow(True)
        ax_histy.set_axisbelow(True)

        if xlabel is not None:
            ax.set_xlabel(xlabel)
        if ylabel is not None:
            ax.set_ylabel(ylabel)

        if filename is not None:
            plt.savefig(filename, dpi=300)
            plt.close()
        else:
            return fig

    def plot_opinions(
        self,
        timestep: Optional[int] = None,
        xlim: Optional[tuple[float, float]] = None,
        ylim: Optional[tuple[float, float]] = None,
        bins: Optional[Union[int, np.ndarray]] = 50,
        print_title: bool = True,
        ax: Optional[plt.Axes] = None,
        density: bool = True,
    ) -> plt.Axes:
        """
        Creates a histogram of the opinion in the population.
        Parameters
        ----------
        @param timestep : int, optional
            timestep/snapshot of the simulation which is to be plotted. The default is None, resulting in plotting of
            the latest snapshot of the simulation.
        @param xlim: Tuple of floats, optional
            interval of the x-axis (lower bound, upper bound). The default is None.
        @param ylim : Tuple of floats, optional
            interval of the y-axis (lower bound, upper bound). The default is None.
        @param bins : Int or ascending array, optional
            number of bins used for the histogram. The default is 50. If an array is supplied, it will define edges of
            the bins.
        @param print_title : Bool, optional
            specifying if the title should be shown. The default is True.
        @param ax : Axes object from matplotlib, optional
            object on which the graph is plotted. The default is None, resulting in the creation of a new Axes object.
        @param measures: Iterable of strings, optional
            Specifies which diversity measures should be added to the legend. It has to be one of the following:
            "average_local_agreement", "polarization", "opinion_diversity", "network_disagreement_index",
            "polarization_level","radicalisation_level"
        @param density : bool
            If True, the histogram will be scaled as a density and not a probability mass function.

        Returns
        -------
        An Axes object from matplotlib containing the plot.

        """

        if ax is None:
            ax = plt.gca()

        if timestep is None:
            timestep = self.model.steps
        opinions = self.model.results["Opinions"][timestep]

        if isinstance(bins, list):
            # ensure we use a numpy array
            bins = np.array(bins)
        if isinstance(bins, np.ndarray):
            # define the bins
            if min(opinions) < bins[0] or max(opinions) > bins[-1]:
                warnings.warn(
                    f"The provided bins do not cover the range of opinions."
                    f" max(Opinions) = {np.round(max(opinions), 4)} and "
                    f"min(opinions) = {np.round(min(opinions), 4)}"
                    f"\n New bins were modified to compensate and bin width remains the same.)"
                )
                bins = np.arange(
                    min(opinions) - 0.0002, max(opinions) + 0.0002, bins[1] - bins[0]
                )

        ax.hist(opinions, bins, density=density)

        if xlim is not None:
            ax.set_xlim(xlim)
        else:
            ax.set_xlim(self.model.opdim)
        if ylim is not None:
            ax.set_ylim(ylim)

        if print_title:
            plt.title(
                "Opinion distibution at time t ={:.2f}".format(
                    timestep * self.model.delta_t
                )
            )

        # plt.tight_layout()
        return ax

    def plot_degree_distributions(
        self,
        ax: Optional[plt.axes] = None,
        plot_title: bool = True,
        time_step: Optional[int] = None,
        bins: Optional[Union[np.ndarray, int]] = None,
    ) -> plt.Axes:
        """
        Plots the degree distribution. If R_sp = R_op the bars are colored according to the average
        opinion of the agents with that degree.

        Parameters
        ---------
        @param ax: list of matplotlib axis objects, optional Default: None
        @param plot_title: bool, optional
             Default is True.
        @param time_step: int, optional
            Timestep to plot the degree distribution
        @param bins: int, ndarray, optional
            If the input is an int, it determines the number of bins, if a nnumpy array is given,
            the entries detrmine the edges of the bins.

        Returns
        ----------
        result:
            matplotlib axis object with a bar plots for the degree distribution
        """
        if ax is None:
            ax = plt.gca()

        if self.model.R_sp != self.model.R_op:
            # If the opinion and spatial radius are different, both degree distributions are plotted.
            opacity = 0.7
            counts, bin_edges = self.model.get_degree_distributions(time_step, bins)[
                "Opinion"
            ]
            ax.bar(
                (bin_edges[:-1] + bin_edges[1:]) / 2,
                counts,
                color="forestgreen",
                alpha=opacity,
                label="Opinion Network",
                width=np.diff(bin_edges),
                edgecolor="k",
            )

            counts, bin_edges = self.model.get_degree_distributions(time_step, bins)[
                "Spatial"
            ]
            ax.bar(
                (bin_edges[:-1] + bin_edges[1:]) / 2,
                counts,
                color="tab:blue",
                alpha=opacity,
                label="Spatial Network",
                width=np.diff(bin_edges),
                edgecolor="k",
            )
            ax.legend()
        else:
            deg_counts, bin_edges = self.model.get_degree_distributions(
                time_step, bins
            )["Spatial"]
            av_opinion = np.zeros(len(bin_edges))
            spatial_network = nx.from_numpy_array(
                self.model.get_adjacency_matrix(useR_sp=True)
            )
            for k, _ in enumerate(bin_edges):
                if k == len(bin_edges):
                    continue
                opinions = []
                for i, agent in enumerate(self.model.agents):
                    if bin_edges[k] < spatial_network.degree[i] < bin_edges[k + 1]:
                        opinions.append(agent.get_opinion())
                av_opinion[k] = np.mean(opinions) if len(opinions) > 0 else 0
            # defining the rescaler, that maps the opinions from [-1,1] to [0,1] () necessary for the colormap
            rescale = lambda y: (y + 1) / 2
            ax.bar(
                (bin_edges[:-1] + bin_edges[1:]) / 2,
                deg_counts,
                color=aux.discontinious_red_blue(rescale(av_opinion)),
                edgecolor="k",
            )

        if plot_title:
            ax.set_title(
                f"Degree distribution \n of induced networks at time {self.model.delta_t * self.model.steps}"
            )
        ax.set_ylabel("Number of Nodes")
        ax.set_ylim((0, self.model.N))
        return ax

    def plot_global_assortativity(
        self, ax: Optional[plt.Axes] = None, use_Rsp=True
    ) -> plt.Axes:
        if ax is None:
            ax = plt.gca()
        time = np.arange(self.model.steps + 1)
        global_assortativity = [
            self.model.global_assortativity(step, use_spatial_radius=use_Rsp)
            for step in time
        ]
        time = self.model.delta_t * time
        ax.plot(time, global_assortativity, label="global assortativity")
        ax.set_xlabel("t")
        ax.set_ylabel("Assortativity")
        ax.grid()

        return ax
