"""
Created by Soeren Nagel

This file contains the ModelStatistics Class that handles any statistics related to the Model class.

For a main overview of how and what data is stored, see the documentation of the
get_results() method.

"""

# standard imports
from __future__ import annotations
import warnings
import time
from typing import Optional

# third party imports
import psutil
import os
import numpy as np
from joblib import Parallel, delayed
from tqdm import tqdm
import pandas as pd
import matplotlib.pyplot as plt


# local Imports
from src import Model


class ModelStatistics:
    """
    This class is ment to handle all statistics and bifurcation analysis of the underlying Model.
    """

    def __init__(
        self,
        num_iterations: int,
        N: int,
        R: float | tuple[float, float],
        alpha: float,
        beta: float,
        sigma_op: float,
        sigma_sp: float,
        delta_t: float,
        init_xlim: tuple[float, float] = (-0.25, 0.25),
        init_ylim: tuple[float, float] = (-0.25, 0.25),
        init_oplim: tuple[float, float] = (-1, 1),
        useDoublewell: bool = False,
        spatial_clustering: bool = True,
        opinion_clustering: bool = True,
        dynamic_eps: bool = False,
        useHDBSCAN: bool = True,
        datasim: bool = False,
        xlim: Optional[tuple[float, float]] = None,
        ylim: Optional[tuple[float, float]] = None,
        boundary: bool = True,
        normalize_with_neighbour_num: bool = False,
        input_df: Optional[pd.DataFrame] | list[pd.DataFrame] = None,
        initialize: bool = True,
        seed: Optional[int] = None,
    ) -> None:
        """
        @param num_iterations: int
            The number of iterations/runs
        @param N: int
            Number of agents in the system.
        @param R: float OR tuple of floats
            minimum radius to facilitate interaction between agents OR tuple defining radius for
             spatial and opinion interaction (opinion radius, spatial radius).
        @param alpha: float
            opinion strength.
        @param beta: float
            spatial strength.
        @param sigma_op: float
            opinion dynamics diffusion coefficient.
        @param sigma_sp: float
            spatial dynamics diffusion coefficient.
        @param delta_t: float
            time between simulation snapshots/steps.
        @param init_xlim: tuple of floats, optional
            defines lower and upper initial x bound (lower bound, upper bound).
            The default is (-0.25,0.25).
        @param init_ylim: tuple of floats, optional
            defines lower and upper initial y bound (lower bound, upper bound).
            The default is (-0.25,0.25).
        init_oplim : tuple of floats, optional
        @param init_oplim: tuple of floats, optional
            defines lower and upper initial opinion bound (lower bound, upper bound).
            The default is (-1,1).
        @param useDoublewell: bool, optional
            specifying if the double well should be used as a potential landscape.
            The default is False.
        @param spatial_clustering: bool, optional
            specifying if the clustering should include spatial coordinates.
            The default is True.
        @param opinion_clustering:  bool, optional
            specifying if the clustering should include opinion coordinates. The default is True.
        @param dynamic_eps: bool, optional
            specifying if eps should be dynamically calculated before each clustering.
            The default is False.
        @param useHDBSCAN: bool, optional
            specifying if HDBSCAN should be used over DBSCAN. The default is True.
        @param datasim: bool, optional
            specifying if data is simulated (adds additional functionalities). The default is False.
        @param xlim: tuple of floats, optional
            defines lower and upper x bounds of the grid (lower bound, upper bound).
            The default is None.
        @param ylim: tuple of floats, optional
            defines lower and upper y bounds of the grid (lower bound, upper bound).
            The default is None.
        @param boundary: bool, optional
            specifying if agents are bounded within their grid cell. The default is True.
        @param normalize_with_neighbour_num: bool, optional
            if True, the interaction terms will be normalized by 1/deg(i) instead of 1/N.
            The default is False.
        @param input_df: Nx3 pd.DataFrame, optional
            Matrix containing x coordinates for all agents in the first column, y coordinates in the
            second column and opinions in the third column. The default is None.
        @param initialize: bool, optional
            determines whether the models should be initialized or not. The default is True.
        @param seed:  int, optional
            seed for generation of random numbers. The default is None.
        """

        # Set Up the parameters
        self.N = N
        if isinstance(
            R, tuple
        ):  # allow tuple input (R_op, R_sp) to define separate values, as well as integer input
            # which assumes R_sp = R_op
            self.R_op = float(R[0])
            self.R_sp = float(R[1])
        else:
            self.R_op = float(R)  # opinion interaction radius
            self.R_sp = float(R)
        self.alpha = float(alpha)
        self.beta = float(beta)
        self.sigma_op = float(sigma_op)
        self.sigma_sp = float(sigma_sp)
        self.delta_t = float(delta_t)
        self.xlim = init_xlim
        self.ylim = init_ylim
        self.oplim = init_oplim
        self.useDoublewell = useDoublewell
        self.spatial_clustering = spatial_clustering
        self.opinion_clustering = opinion_clustering
        self.dynamic_eps = dynamic_eps
        self.useHDBSCAN = useHDBSCAN
        self.datasim = datasim
        self.xlim = xlim
        self.ylim = ylim
        self.boundary = boundary
        self.normalize_with_neighbour_num = normalize_with_neighbour_num
        self.seed = seed
        self.num_iterations = num_iterations

        self.models = [
            Model(
                N=N,
                R=R,
                alpha=alpha,
                beta=beta,
                sigma_op=sigma_op,
                sigma_sp=sigma_sp,
                delta_t=delta_t,
                init_xlim=init_xlim,
                init_ylim=init_ylim,
                init_oplim=init_oplim,
                useDoublewell=useDoublewell,
                seed=seed,
                spatial_clustering=spatial_clustering,
                opinion_clustering=opinion_clustering,
                dynamic_eps=dynamic_eps,
                useHDBSCAN=useHDBSCAN,
                datasim=datasim,
                xlim=xlim,
                ylim=ylim,
                boundary=boundary,
                normalize_with_neighbour_num=normalize_with_neighbour_num,
            )
            for _ in range(num_iterations)
        ]
        if initialize:
            if isinstance(input_df, list):
                for model, df in zip(self.models, input_df):
                    model.initialize_model(input_df=df)
            else:
                for model in self.models:
                    model.initialize_model(input_df=input_df)

        self.steps = self.models[0].steps
        self.data_is_saved = (
            False  # Will be used to keep track of whether we loaded or simulated
        )
        # the data. In order not to save the same runs multiple times in the databank

    def __repr__(self):
        """Returns a string that represents the class instance. This is used to store data and load
        data. This representation can be used to create a new instance of the class."""
        # I have some data stored, where these are stored as integers.
        # These would not be found otherwise.
        # This is not necessary otherwise.
        if isinstance(self.alpha, float):
            if self.alpha.is_integer():
                self.alpha = int(self.alpha)
        if isinstance(self.beta, float):
            if self.beta.is_integer():
                self.beta = int(self.beta)

        return (
            f"ModelStatistics({self.num_iterations}, {self.N}, {self.R_op}, {self.R_sp}, {self.alpha}, {self.beta}, {self.sigma_op}, {self.sigma_sp}, "
            f"{self.delta_t}, {str(self.xlim)}, {str(self.ylim)}, {str(self.oplim)}, "
            f"{self.useDoublewell}, {self.spatial_clustering}, {self.opinion_clustering}, {self.dynamic_eps},"
            f"{self.useHDBSCAN}, {self.datasim}, {self.boundary}, {self.normalize_with_neighbour_num}, "
            f"{self.steps}, {self.seed})"
        )

    def __eq__(self, other):
        """
        Compares all the relevant data, to check if other contains runs of the same model as self.
        Note, that this will not check if the models have the same initial data.

        @param other: another instance of the ModelStatistics class
        """
        test = (
            self.N == other.N,
            self.R_op == other.R_op,
            self.R_sp == other.R_sp,
            self.alpha == other.alpha,
            self.beta == other.beta,
            self.sigma_op == other.sigma_op,
            self.sigma_sp == other.sigma_sp,
            self.delta_t == other.delta_t,
            self.xlim == other.xlim,
            self.ylim == other.ylim,
            self.oplim == other.oplim,
            self.useDoublewell == other.useDoublewell,
            self.spatial_clustering == other.spatial_clustering,
            self.opinion_clustering == other.opinion_clustering,
            self.dynamic_eps == other.dynamic_eps,
            self.useHDBSCAN == other.useHDBSCAN,
            self.datasim == other.datasim,
            self.boundary == other.boundary,
            self.normalize_with_neighbour_num == other.normalize_with_neighbour_num,
            self.steps == other.steps,
            self.data_is_saved == other.data_is_saved,
            self.seed == other.seed,
        )

        if all(test):
            return True
        else:
            for i, val in enumerate(test):
                if not val:
                    print(f"{i}-th value is false")
            return False

    def __add__(self, other) -> ModelStatistics:
        """
        Addition should be used on class instances to combine statistics. If we have two
        instances with identical parameters and the same number of steps, we want to combine the
        simulation results into a larger class, containing all the results.
        """
        if self == other:
            # sets up the new Class instance with updated num_iterations
            result = ModelStatistics(
                self.num_iterations + other.num_iterations,
                self.N,
                (self.R_op, self.R_sp),
                self.alpha,
                self.beta,
                self.sigma_op,
                self.sigma_sp,
                self.delta_t,
                self.xlim,
                self.ylim,
                self.oplim,
                self.useDoublewell,
                self.spatial_clustering,
                self.opinion_clustering,
                self.dynamic_eps,
                self.useHDBSCAN,
                self.datasim,
                self.xlim,
                self.ylim,
                self.boundary,
                self.normalize_with_neighbour_num,
                input_df=None,
                initialize=False,  # The model are already initialized in general.
            )
            # updates the list of models containing the simulation data
            result.models = self.models + other.models
            return result
        else:
            raise ValueError(
                "Then instances you are trying to combine don't come from identical models"
            )

    # ##############################################################################################
    # 1) Data Management
    # ##############################################################################################
    def get_results(self) -> dict:
        """
        Returns the dictionary containing the simulation data from all the individual
        Model instances.

        Returns
        -------
        dictionary
            Dictionary containing "Coordinates",
            "Opinions"

            "Coordinates": lists of pandas DataFrames
                Each DataFrame
                (i.e., get_results()[time_step]) contains the x coordinates of
                agents in the first column and y coordinates of agents in the second column.
                The
                list index corresponds to the simulation snapshot.
                The coordinaters are stacked on
                top of each other fo the columns of each dataframe have length
                slef.num_iterations*self.N. To access the data of indivual models, use the
                get_results on individual models in self.models.


            "Opinions": pandas DataFrame
                Rows represent agents and columns represent simulation snapshots.
                The entries are the opinion values.
                The shape of the dataframe is (num_agents x num_iterations) x num_timesteps
        """
        results = {}
        results["Coordinates"] = [
            # Stacking all the coordinates for a single timestep ontop of each other (axis=1)
            #
            pd.concat(
                [model.get_results()["Coordinates"][i] for model in self.models],
                axis=0,
                ignore_index=True,
            )
            for i, _ in enumerate(self.models[0].get_results()["Coordinates"])
        ]

        results["Opinions"] = pd.concat(
            [model.get_results()["Opinions"] for model in self.models],
            axis=0,
            ignore_index=True,
        )
        return results

    def simulate(self, steps: int, n_processes=-1) -> None:
        """
        Runs the simulations in parallel to speed up thinks a little bit.

        Parameters
        ----------
        steps: int Number of steps for each simulation
        n_processes: int Number of processes to be started simultaneously Default: Use all available
        """
        if self.data_is_saved:
            raise ValueError(
                "Would modify saved data. Start from empty Model, when simulating."
            )

        def run_sim(model, iterations: int) -> None:
            """Just a placeholder function, that is passed to joblib.Parallel"""
            model.simulate(iterations)
            return model

        # run everything in parallel and update the model and get some data on memory usage
        process = psutil.Process()
        start_time = time.time()
        start_memory = process.memory_info().rss
        self.models = Parallel(n_jobs=n_processes)(
            delayed(run_sim)(model, steps)
            for model in tqdm(self.models, desc="Integrating SDEs")
        )
        end_time = time.time()
        end_memory = process.memory_info().rss
        elapsed_time = end_time - start_time
        memory_usage = end_memory - start_memory

        print(
            f"Time per parameter: {elapsed_time} seconds,"
            f" Memory Usage: {memory_usage / (1024 ** 3)} GB"
        )

        # Just to check that noting went wrong
        if len(set([model.steps for model in self.models])) != 1:
            raise ValueError(f"The models were not integrated for the same time.")

        self.steps = self.models[0].steps  # Updating the steps
        self.data_is_saved = "simulated"  # Keeping simulated and loaded data separate

    def save_results_as_numpy(self, folder: str, final=False) -> None:
        """
        Save the results to a .npy file in the specified folder.
        The name is python-readable (without the .npy) file extension.
        So any stored data can be read by creating a new instance of the class by using the
        filename and then loading the data.
        @param folder:
        @return: None
        """
        if self.data_is_saved != "simulated":
            raise ValueError(
                "This data is either loaded from existing file or not simulated data and can't be "
                "saved."
            )
        file = f"{folder}/" + self.__repr__() + f".npy"
        if final:
            file = f"{folder}/" + self.__repr__() + f"_final.npy"
        data = []
        for i, model in tqdm(
            enumerate(self.models), desc="converting results to numpy array"
        ):
            data.append(model.get_results_as_numpy_array())
        np.save(file, np.array(data))
        print("saving complete")

    def data_exists(self, steps, final=False) -> bool:
        """Checks if this instance exists in the DataBank folder.
        The steps need to be specified, since data can only be loaded into an empty instance.
        """
        self.steps = steps
        base_dir = os.path.dirname(__file__)
        data_dir = os.path.join(base_dir, "..", "data", "DataBank")
        if final:
            file_path = os.path.join(data_dir, repr(self) + f"_final.npy")
        else:
            file_path = os.path.join(data_dir, repr(self) + f".npy")
        if os.path.isfile(file_path):
            return True
        self.steps = 0
        return False

    def load_results_from_numpy(self, folder, final=False) -> None:
        """Loads saved data from the given folder.
        The filename is automatically defined via the __repr__ method.
        So any data can be read by creating a new instance of the class by using the filename
        without the .npy extension."""
        if self.data_is_saved or self.data_is_saved == "simulated":
            raise ValueError("This instance contains either simulated or loaded data.")
        self.data_is_saved = True
        if final:
            file = f"{folder}/" + self.__repr__() + f"_final.npy"
        else:
            file = f"{folder}/" + self.__repr__() + f".npy"

        data = np.load(file)

        for results, empty_model in tqdm(
            zip(data, self.models), desc="Converting Models"
        ):
            empty_model.set_results_from_numpy_array(results)
            empty_model.steps = self.steps

    def clear(self) -> None:
        """
        Clears any data in the instance and creates new empty model instances.
        This method can be used if the amounts of data exceed the memory.
        """
        self.models = [
            Model(
                self.N,
                (self.R_op, self.R_sp),
                self.alpha,
                self.beta,
                self.sigma_op,
                self.sigma_sp,
                self.delta_t,
                self.xlim,
                self.ylim,
                self.oplim,
                self.useDoublewell,
                self.seed,
                self.spatial_clustering,
                self.opinion_clustering,
                self.dynamic_eps,
                self.useHDBSCAN,
                self.datasim,
                self.xlim,
                self.ylim,
                self.boundary,
                self.normalize_with_neighbour_num,
            )
            for _ in range(self.num_iterations)
        ]
        for model in self.models:
            model.steps = self.steps
        self.data_is_saved = False

    def global_assortativity(self, time_step=None) -> np.ndarray:
        # calculate all the assortativities for all timesteps
        if time_step is None:
            assortativity = np.zeros((self.num_iterations, self.steps + 1))
            for i, model in tqdm(
                enumerate(self.models), desc="Calculating Assortativity"
            ):
                assortativity[i, :] = [
                    model.global_assortativity(step) for step in range(self.steps + 1)
                ]
            return np.array(assortativity)
        # calculate all the assortativities for a single timestep
        else:
            assortativity = np.zeros(self.num_iterations)
            for i, model in enumerate(self.models):
                assortativity[i] = model.global_assortativity(time_step=time_step)
            return np.array(assortativity)

    def plot_opinions(
        self,
        timestep: Optional[int] = None,
        ylim: Optional[tuple[float, float]] = None,
        bins: int | np.ndarray = 30,
        print_title: bool = True,
        ax: Optional[plt.Axes] = None,
    ) -> plt.Axes:
        """
        Plots the average opinionn distribution for the given timestep.

        Parameters
        ----------
        @param timestep : int, optional
            timestep/snapshot of the simulation which is to be plotted. The default is None,
            resulting in plotting of the latest snapshot of the simulation.
        @param ylim : Tuple of floats, optional
            interval of the y-axis (lower bound, upper bound). The default is None.
        @param bins : int or ascending array, optional
            number of bins used for the histogram. The default is 50. If an array is supplied, it
            will define edges of the bins
        @param print_title : bool, optional
            specifying if the title should be shown. The default is True.
        @param ax : Axes object from matplotlib, optional
            object on which the graph is plotted. The default is None, resulting in the creation of
            a new Axes object.

        Returns
        -------
        An Axes object from matplotlib containing the plot.

        """

        if ax is None:
            ax = plt.gca()

        if timestep is None:
            timestep = self.steps
        opinions = self.get_results()["Opinions"][timestep]
        if isinstance(bins, list):
            bins = np.array(bins)
        if isinstance(bins, np.ndarray):
            if min(opinions) < bins[0] or max(opinions) > bins[-1]:
                warnings.warn(f"The provided bins do not cover the range of opinions.")

        hist, bins = np.histogram(opinions, bins, density=True)

        # Plot the resulting histogram
        center = (bins[:-1] + bins[1:]) / 2
        ax.bar(center, hist, align="center", width=bins[1] - bins[0])
        ax.set_xlim(self.oplim)

        if ylim is not None:
            ax.set_ylim(ylim)

        if print_title:
            plt.title(
                f"Opinion distribution at time t = {timestep * self.delta_t} averaged for "
                f"{self.num_iterations} runs\n"
                f"for $N ={self.N}, R_{{op}} ={self.R_op}, R_{{sp}} ={self.R_sp}, "
                f"\\alpha ={self.alpha}, \\beta ={self.beta}, \\sigma_{{op}} ={self.sigma_op}, "
                f"\\sigma_{{sp}} ={self.sigma_sp}$"
            )

        # plt.tight_layout()
        ax.set_xlabel(r"$\Theta$")
        ax.set_ylabel(r"Density $P(\Theta)$")
        ax.grid(True)

        return ax


if __name__ == "__main__":
    # model parameters for a base model
    N = 10
    R = 0.15
    alpha = 20
    beta = 20
    sigma_op = 0.05
    sigma_sp = 0.05
    delta_t = 0.01
    local_coupling = True
    KDE_bandwidth = 0.02
    timestep = 25

    # simulation parameters
    n_iter = 7
    num_processes = 7
    # plotting_parameters -> For all figures
    deg_lim = (0, 0.26)
    opinion_ylim = (0, 0.2)
    xlim = (-0.5, 0.5)
    ylim = (-0.5, 0.5)

    bins = 100
    folder = "Images/Bin"

    # =============================================================================
    # base case
    # =============================================================================

    test_model = ModelStatistics(
        n_iter,
        N,
        R,
        alpha,
        beta,
        sigma_op,
        sigma_sp,
        delta_t,
        normalize_with_neighbour_num=local_coupling,
    )
    test_model.simulate(20)

    res = test_model.get_results()
    fig, ax = plt.subplots()
    test_model.plot()
    plt.show()
