import unittest
import numpy as np
import pandas as pd

from src import Model


class TestModel(unittest.TestCase):

    def setUp(self):
        self.threhold_global_asssortativity = 10e-17
        self.threshold_multi_scale_asssortativity = 10e-10
        self.threhold_kl_divergence = 10e-10
        self.seed = 233
        self.rng = np.random.default_rng(self.seed)
        self.model = Model(
            N=100,
            R=0.15,
            alpha=20,
            beta=20,
            sigma_op=0.001,
            sigma_sp=0.001,
            delta_t=0.01,
            datasim=False,
        )

    def test_interaction(self):
        """
        This will test if the new implementation of the dynamics coincides with the old implementation.
        The test evaluates the RHS of the dynamics for random initial condition and after a single integration step.
        At some point we need to save a reference, so that we can remove the old implementation.
        """
        x = self.rng.normal(0, 0.01, 100)
        y = self.rng.normal(0, 0.01, 100)
        opinions = self.rng.normal(0, 1, 100)
        input_df = pd.DataFrame({"x": x, "y": y, "opinions": opinions})

        self.model.initialize_model(input_df)

        old_rhs = np.zeros(3 * self.model.N)
        for i in range(self.model.N):
            # get current data_is_saved
            old_rhs[i], old_rhs[i + self.model.N] = self.model.spatial_interaction(i)
            old_rhs[i + 2 * self.model.N] = self.model.opinion_interaction(i)

        initial_data = np.concatenate(
            [
                input_df["x"].to_numpy(),
                input_df["y"].to_numpy(),
                input_df["opinions"].to_numpy(),
            ]
        )
        new_rhs = self.model.interactions(None, initial_data, None)
        # chek if all the enties of  the old and new implementation agree to the 10th decimal place
        for old_val, new_val in zip(old_rhs, new_rhs):
            self.assertLessEqual(np.abs(old_val - new_val), 10**-5)

        # Now we check that taking an integration step does not brake anything
        self.model.simulate()
        current_data = np.concatenate(
            [
                self.model.get_results()["Coordinates"][-1].to_numpy()[:, 0],
                self.model.get_results()["Coordinates"][-1].to_numpy()[:, 1],
                self.model.get_results()["Opinions"][1].to_numpy(),
            ]
        )
        new_rhs = self.model.interactions(None, current_data, None)
        for i in range(self.model.N):
            # get current data_is_saved
            old_rhs[i], old_rhs[i + self.model.N] = self.model.spatial_interaction(i)
            old_rhs[i + 2 * self.model.N] = self.model.opinion_interaction(i)

        # check if all the enties of  the old and new implementation agree to the 10th decimal place
        for old_val, new_val in zip(old_rhs, new_rhs):
            self.assertLessEqual(np.abs(old_val - new_val), 10**-5)

    def test_integration(self):
        """
        This will test, if our model will converge to the mean opinion and mean position of the random initial data.
        This is the case without noise, large enough interaction radii and strictly positive opinions
        """
        x = self.rng.random(self.model.N)
        y = self.rng.random(self.model.N)
        # create initial conditions that avoid the repulsive interactions
        # this gives us a standard three dimensional deGroot model
        opinions = self.rng.random(self.model.N) + 0.2
        assert opinions.min() > 0
        input_df = pd.DataFrame({"x": x, "y": y, "opinions": opinions})

        # Ensure the network will remain fully connected
        self.model.R_sp = 50
        self.model.R_op = 50

        # faster convergence
        self.model.alpha = 50
        self.model.beta = 50

        # remove noise
        self.model.sigma_op = 0
        self.model.sigma_sp = 0

        # initialize and simulate
        self.model.initialize_model(input_df)
        self.model.simulate(400)

        for step in range(self.model.steps):
            final_x, final_y = (
                self.model.get_results()["Coordinates"][step].to_numpy().transpose()
            )
            final_opinions = self.model.get_results()["Opinions"][step]

            # Check if the mean is a conserved quantity throughout the integration
            self.assertLessEqual(
                final_x.mean() - x.mean(),
                10**-2,
                msg=f"Integrator lost accuracy after {step} steps",
            )
            self.assertLessEqual(final_y.mean() - y.mean(), 10**-2)
            self.assertLessEqual(final_opinions.mean() - opinions.mean(), 10**-2)

        # Test convergence in x
        mean_distance_consensus = np.sqrt(
            np.mean((final_opinions - opinions.mean()) ** 2)
        )
        self.assertLessEqual(mean_distance_consensus, 10**-2)

        # Test convergence in y
        mean_distance_consensus = np.sqrt(np.mean((final_x - x.mean()) ** 2))
        self.assertLessEqual(mean_distance_consensus, 10**-2)

        # test convergence in opinions
        mean_distance_consensus = np.sqrt(np.mean((final_y - y.mean()) ** 2))
        self.assertLessEqual(mean_distance_consensus, 10**-2)

    def test_new_integration(self):
        """
        This will test, if our model will converge to the mean opinion and mean position of the random initial data.
        This is the case without noise, large enough interaction radii and strictly positive opinions
        """
        x = self.rng.random(self.model.N)
        y = self.rng.random(self.model.N)
        # create initial conditions that avoid the repulsive interactions
        # this gives us a standard three dimensional deGroot model
        opinions = self.rng.random(self.model.N) + 0.2
        assert opinions.min() > 0
        input_df = pd.DataFrame({"x": x, "y": y, "opinions": opinions})

        # Ensure the network will remain fully connected
        self.model.R_sp = 50
        self.model.R_op = 50

        # faster convergence
        self.model.alpha = 80
        self.model.beta = 80

        # remove noise
        self.model.sigma_op = 0
        self.model.sigma_sp = 0

        # clustering is not implemented
        self.model.clustering = False
        self.model.opinion_clustering = False
        self.model.spatial_clustering = False

        # initialize and simulate
        self.model.initialize_model(input_df)
        self.model.simulate_new(400)

        for step in range(self.model.steps):
            final_x, final_y = (
                self.model.get_results()["Coordinates"][step].to_numpy().transpose()
            )
            final_opinions = self.model.get_results()["Opinions"][step]

            # Check if the mean is a conserved quantity throughout the integration
            self.assertLessEqual(
                final_x.mean() - x.mean(),
                10**-2,
                msg=f"Integrator lost accuracy after {step} steps",
            )
            self.assertLessEqual(final_y.mean() - y.mean(), 10**-2)
            self.assertLessEqual(final_opinions.mean() - opinions.mean(), 10**-2)

        final_x, final_y = (
            self.model.get_results()["Coordinates"][self.model.steps]
            .to_numpy()
            .transpose()
        )
        final_opinions = self.model.get_results()["Opinions"][
            self.model.steps
        ].to_numpy()

        # Test convergence in opinions
        mean_distance_consensus = np.sqrt(
            np.mean((final_opinions - final_opinions.mean()) ** 2)
        )
        self.assertLessEqual(mean_distance_consensus, 10**-2)

        # Test convergence in y
        mean_distance_consensus = np.sqrt(np.mean((final_x - final_x.mean()) ** 2))
        self.assertLessEqual(mean_distance_consensus, 10**-2)

        # test convergence in opinions
        mean_distance_consensus = np.sqrt(np.mean((final_y - final_y.mean()) ** 2))
        self.assertLessEqual(mean_distance_consensus, 10**-2)
